#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import locale
import os
import re
from typing import Any, Dict, List, Optional, Tuple

import oss2
from alibabacloud_credentials.client import Client as CredentialClient
from alibabacloud_credentials.models import Config as CredentialConfig
from alibabacloud_sts20150401.client import Client
from alibabacloud_sts20150401.models import (
    GetCallerIdentityResponseBody as CallerIdentity,
)
from alibabacloud_tea_openapi import models as open_api_models
from oss2.models import BucketInfo, SimplifiedBucketInfo
from prompt_toolkit import Application
from prompt_toolkit.key_binding import KeyBindings, merge_key_bindings
from prompt_toolkit.key_binding.defaults import load_key_bindings
from prompt_toolkit.layout import HSplit, Layout
from prompt_toolkit.shortcuts import confirm as prompt_confirm
from prompt_toolkit.widgets import Label, RadioList

from ...api.base import ServiceName
from ...api.client_factory import ClientFactory
from ...api.workspace import WorkspaceAPI, WorkspaceConfigKeys
from ...common.consts import DEFAULT_NETWORK_TYPE, PAI_VPC_ENDPOINT, Network
from ...common.logging import get_logger
from ...common.oss_utils import CredentialProviderWrapper, OssUriObj
from ...common.utils import is_domain_connectable, make_list_resource_iterator
from ...libs.alibabacloud_pai_dsw20220101.client import Client as DswClient
from ...libs.alibabacloud_pai_dsw20220101.models import GetInstanceRequest
from ...session import Session

logger = get_logger(__name__)

locale_code, _ = locale.getdefaultlocale()

OSS_NAME_PATTERN = re.compile(pattern="^[a-z0-9][a-z0-9-]{1,61}[a-z0-9]$")
ZH_CN_LOCAL = "zh_CN"

# RoleARN pattern for AssumedRole CallerIdentity
ASSUMED_ROLE_ARN_PATTERN = re.compile(r"acs:ram::\d+:assumed-role/([^/]+)/.*")

# DSW Notebook Default Role Name:
PAI_DSW_DEFAULT_ROLE_NAME = "aliyunpaidswdefaultrole"


DEFAULT_PRODUCT_RAM_ROLE_NAMES = [
    "AliyunODPSPAIDefaultRole",
    "AliyunPAIAccessingOSSRole",
    "AliyunPAIDLCAccessingOSSRole",
    "AliyunPAIDLCDefaultRole",
]


class WorkspaceRoles(object):
    """Workspace roles."""

    AlgoDeveloper = "PAI.AlgoDeveloper"
    WorkspaceAdmin = "PAI.WorkspaceAdmin"
    WorkspaceOwner = "PAI.WorkspaceOwner"
    LabelManager = "PAI.LabelManager"

    @classmethod
    def recommend_roles(cls):
        """Recommend roles for user to use."""
        return [
            cls.AlgoDeveloper,
            cls.WorkspaceAdmin,
            cls.WorkspaceOwner,
        ]


class CallerIdentityType(object):

    # Document: https://help.aliyun.com/document_detail/371868.html

    # - Account: an Alibaba Cloud account
    Account = "Account"
    # - RamUser: a RAM user
    RamUser = "RAMUser"
    # - AssumedRoleUser: a RAM role
    AssumedRoleUser = "AssumedRoleUser"


class UserProfile(object):

    _credential_client = None

    def __init__(
        self,
        credential_config: CredentialConfig,
        region_id: str,
    ):
        self.region_id = region_id
        self.credential_config = credential_config

        if DEFAULT_NETWORK_TYPE:
            self.network = Network.from_string(DEFAULT_NETWORK_TYPE)
        else:
            self.network = (
                Network.VPC
                if is_domain_connectable(PAI_VPC_ENDPOINT.format(self.region_id))
                else Network.PUBLIC
            )
        self._caller_identify = self._get_caller_identity()

    def _get_credential_client(self):
        if self._credential_client:
            return self._credential_client
        self._credential_client = CredentialClient(self.credential_config)
        return self._credential_client

    def get_access_key_id(self):
        return self._get_credential_client().get_access_key_id()

    def get_access_key_secret(self):
        return self._get_credential_client().get_access_key_secret()

    def get_security_token(self):
        return self._get_credential_client().get_security_token()

    def _get_caller_identity(self) -> CallerIdentity:
        return (
            Client(
                config=open_api_models.Config(
                    credential=self._get_credential_client(),
                    region_id=self.region_id,
                    network=(
                        None
                        if self.network == Network.PUBLIC
                        else self.network.value.lower()
                    ),
                )
            )
            .get_caller_identity()
            .body
        )

    def is_dsw_default_role(self) -> bool:
        if self._caller_identify.identity_type != CallerIdentityType.AssumedRoleUser:
            return False
        m = ASSUMED_ROLE_ARN_PATTERN.match(self._caller_identify.arn)
        return m and m.group(1).lower() == PAI_DSW_DEFAULT_ROLE_NAME

    def get_acs_dsw_client(self) -> DswClient:
        return ClientFactory.create_client(
            service_name=ServiceName.PAI_DSW,
            credential_client=self._get_credential_client(),
            region_id=self.region_id,
            network=self.network,
        )

    def get_instance_info(self, instance_id: str) -> Dict[str, Any]:
        dsw_client = self.get_acs_dsw_client()
        return dsw_client.get_instance(
            instance_id, request=GetInstanceRequest()
        ).body.to_map()

    def get_credential(self):
        return self._credential_client.get_access_key_id()

    @property
    def is_ram_user(self) -> bool:
        return self._caller_identify.identity_type == CallerIdentityType.RamUser

    @property
    def is_account(self) -> bool:
        return self._caller_identify.identity_type == CallerIdentityType.Account

    @property
    def account_id(self):
        """Return Alibaba Cloud account ID of the current user profile"""
        return self._caller_identify.account_id

    @property
    def user_id(self):
        """Return the Alibaba Cloud user ID of the current user profile"""
        return self._caller_identify.user_id

    @property
    def identify_type(self):
        return self._caller_identify.identity_type

    def get_default_oss_endpoint(self):
        return "https://oss-{}.aliyuncs.com".format(self.region_id)

    def list_oss_buckets(self, prefix: str = "") -> List[SimplifiedBucketInfo]:
        buckets: List[SimplifiedBucketInfo] = []
        service = oss2.Service(
            auth=oss2.ProviderAuth(
                credentials_provider=CredentialProviderWrapper(
                    config=self.credential_config,
                ),
            ),
            endpoint=self.get_default_oss_endpoint(),
        )

        marker = ""
        while True:
            res: oss2.models.ListBucketsResult = service.list_buckets(
                prefix=prefix, marker=marker
            )
            buckets.extend(
                [b for b in res.buckets if self.region_id in b.location] or []
            )
            if not res.is_truncated:
                break
            else:
                marker = res.next_marker

        return buckets

    def get_bucket_info(self, bucket_name) -> BucketInfo:
        auth = oss2.ProviderAuth(
            credentials_provider=CredentialProviderWrapper(
                config=self.credential_config,
            ),
        )
        bucket = oss2.Bucket(
            auth, self.get_default_oss_endpoint(), bucket_name=bucket_name
        )
        bucket_info = bucket.get_bucket_info()
        return bucket_info

    def create_oss_bucket(self, bucket_name):
        bucket = oss2.Bucket(
            bucket_name=bucket_name,
            auth=oss2.ProviderAuth(
                credentials_provider=CredentialProviderWrapper(
                    config=self.credential_config,
                ),
            ),
            endpoint=self.get_default_oss_endpoint(),
        )
        bucket.create_bucket()

    def get_production_authorizations(self):
        workspace_api = self.get_workspace_api()
        res = workspace_api.list_product_authorizations(
            ram_role_names=DEFAULT_PRODUCT_RAM_ROLE_NAMES
        )
        return res["AuthorizationDetails"]

    def get_workspace_api(self) -> WorkspaceAPI:
        acs_ws_client = ClientFactory.create_client(
            service_name=ServiceName.PAI_WORKSPACE,
            credential_client=self._get_credential_client(),
            region_id=self.region_id,
            network=self.network,
        )

        return WorkspaceAPI(
            acs_client=acs_ws_client,
        )

    def get_default_oss_storage_uri(
        self, workspace_id: str
    ) -> Tuple[Optional[str], Optional[str]]:
        bucket_name, endpoint = Session.get_default_oss_storage(
            workspace_id=workspace_id,
            cred=self._get_credential_client(),
            region_id=self.region_id,
            network=self.network,
        )
        return "oss://{}/".format(bucket_name), endpoint

    def set_default_oss_storage(
        self, workspace_id, bucket_name: str, intranet_endpoint: str
    ):
        workspace_api = self.get_workspace_api()
        oss_uri = "oss://{}.{}/".format(bucket_name, intranet_endpoint)
        configs = {WorkspaceConfigKeys.DEFAULT_OSS_STORAGE_URI: oss_uri}
        workspace_api.update_configs(workspace_id, configs=configs)

    def get_roles_in_workspace(
        self, workspace_id, user_id: Optional[str] = None
    ) -> List[str]:
        workspace_api = self.get_workspace_api()
        user_id = user_id or self.user_id
        member_info = next(
            (
                mem
                for mem in make_list_resource_iterator(
                    workspace_api.list_members,
                    workspace_id=workspace_id,
                )
                if mem["UserId"] == user_id
            ),
            None,
        )

        # If user has PAIFullAccess policy, 'member_info' may be None.
        return member_info["Roles"] if member_info else []

    def has_permission_edit_config(self, workspace_id: str) -> bool:
        """Return True if the current user has permission to edit workspace config.

        Only members with the role of WorkspaceAdmin or WorkspaceOwner can edit
        workspace config.

        """
        roles = self.get_roles_in_workspace(workspace_id)
        return any(
            (
                r in roles
                for r in [WorkspaceRoles.WorkspaceAdmin, WorkspaceRoles.WorkspaceOwner]
            )
        )


def localized_text(en_text: str, cn_text: str = None):
    if locale_code == ZH_CN_LOCAL:
        return cn_text or en_text
    else:
        return en_text


def mask_secret(secret, mask_count=4):
    masked = (
        secret[:mask_count]
        + (len(secret) - mask_count * 2) * "*"
        + secret[-mask_count:]
    )
    return masked


def mask_and_trim(secret, max_size=20):
    masked = mask_secret(secret, mask_count=8)
    if len(masked) > max_size:
        masked = masked[:max_size] + "..."
    return masked


def radio_list_prompt(
    title: str = "",
    values=None,
    cancel_value=None,
    style=None,
    async_: bool = False,
    **kwargs,
):
    # Create the radio list
    radio_list = RadioList(values)
    # Remove the enter key binding so that we can augment it
    radio_list.control.key_bindings.remove("enter")

    bindings = KeyBindings()

    # Replace the enter key binding to select the value and also submit it
    @bindings.add("enter")
    def exit_with_value(event):
        """
        Pressing Enter will exit the user interface, returning the highlighted value.
        """
        radio_list._handle_enter()
        event.app.exit(result=radio_list.current_value)

    @bindings.add("c-c")
    def backup_exit_with_value(event):
        """
        Pressing Ctrl-C will exit the user interface with the cancel_value.
        """
        event.app.exit(result=cancel_value)

    # Create and run the mini inline application
    application = Application(
        layout=Layout(HSplit([Label(title), radio_list])),
        key_bindings=merge_key_bindings([load_key_bindings(), bindings]),
        mouse_support=True,
        style=style,
        **kwargs,
    )
    if async_:
        return application.run_async()
    else:
        return application.run()


def confirm(message: str = "Confirm?", suffix: str = " (y/n, default: y)"):
    # Input enter key returns an empty string, we assume enter is 'YES'.
    res = prompt_confirm(message, suffix)
    yes = True if isinstance(res, str) and res.strip() == "" else res
    return yes


def not_empty(text: str) -> bool:
    return bool(text.strip())


def print_highlight(msg: str):
    print(ColorEscape.green(msg))


def print_warning(msg: str):
    print(ColorEscape.red(msg))


def validate_bucket_name(name: str) -> bool:
    return bool(OSS_NAME_PATTERN.match(name))


class ColorEscape(object):
    """
    A utility class to wrap a string with color escape code.
    """

    _black = "\u001b[30m"
    _red = "\u001b[31m"
    _green = "\u001b[32m"
    _yellow = "\u001b[33m"
    _blue = "\u001b[34m"
    _magenta = "\u001b[35m"
    _cyan = "\u001b[36m"
    _white = "\u001b[37m"
    _default = "\u001b[39m"

    _reset = "\u001b[0m"

    @classmethod
    def green(cls, msg: str) -> str:
        return cls._format(msg, cls._green)

    @classmethod
    def red(cls, msg: str) -> str:
        return cls._format(msg, cls._red)

    @classmethod
    def _format(cls, msg: str, code: str) -> str:
        if os.environ.get("NO_COLOR"):
            # See https://no-color.org/
            return msg
        return code + msg + cls._reset



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
import subprocess
import time
from random import randint
from typing import Any, Dict, List, Optional, Union

from .logging import get_logger

logger = get_logger(__name__)


def _run_command(command: List[str], input: Optional[str] = None):
    with subprocess.Popen(
        command,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=False,
        bufsize=1,
    ) as p:
        if input:
            p.stdin.write(input.encode())
        p.stdin.close()
        out = io.TextIOWrapper(p.stdout, newline="", errors="replace")
        for line in out:
            logger.info(line)

    return p.returncode


class ContainerRun(object):
    """A class represent a container run in local."""

    CONTAINER_STATUS_RUNNING = "running"
    CONTAINER_STATUS_EXITED = "exited"
    CONTAINER_STATUS_PAUSED = "paused"

    def __init__(self, container, port: Optional[int] = None):
        """Initialize a container run.

        Args:
            container: A docker container object.
            port (int): The host port that container is exposed to.

        """
        self.container = container
        self.port = port

    @property
    def status(self):
        self.container.reload()
        return self.container.status

    def is_running(self):
        """Return True if container is running, otherwise False."""
        return self.status == self.CONTAINER_STATUS_RUNNING

    def is_terminated(self):
        """Return True if container is terminated, otherwise False."""
        return self.status in [
            self.CONTAINER_STATUS_EXITED,
            self.CONTAINER_STATUS_PAUSED,
        ]

    def is_succeeded(self):
        """Return True if container is succeeded, otherwise False."""
        return (
            self.status == "exited" and self.container.attrs["State"]["ExitCode"] == 0
        )

    def wait_for_ready(self, interval=5):
        """Wait until container enter running state or terminated state."""
        while True:
            status = self.status
            if status == self.CONTAINER_STATUS_RUNNING:
                break
            elif status in [self.CONTAINER_STATUS_EXITED, self.CONTAINER_STATUS_PAUSED]:
                raise RuntimeError(
                    "Container is terminated : id={} status={}".format(
                        self.container.id, self.container.status
                    )
                )
            time.sleep(interval)

    def stop(self):
        if self.is_running():
            self.container.stop()

    def start(self):
        if not self.is_running():
            self.container.start()

    def delete(self):
        if self.is_running():
            self.container.stop()
        self.container.remove()

    def watch(self, show_logs: bool = True):
        """Watch container log and wait for container to exit."""
        if not show_logs:
            self.container.wait()
        else:
            log_iter = self.container.logs(
                stream=True,
                follow=True,
            )
            for log in log_iter:
                print(log.decode())

        self.container.reload()
        exit_code = self.container.attrs["State"]["ExitCode"]
        if exit_code != 0:
            raise RuntimeError(
                "Container run exited failed: exit_code={}".format(exit_code)
            )


def run_container(
    image_uri: str,
    container_name: Optional[str] = None,
    port: Optional[int] = None,
    environment_variables: Optional[Dict[str, str]] = None,
    command: Optional[Union[List[str], str]] = None,
    entry_point: Optional[Union[List[str], str]] = None,
    volumes: Optional[Dict[str, Any]] = None,
    working_dir: Optional[str] = None,
    gpu_count: Optional[int] = None,
    gpu_device_ids: Optional[List[str]] = None,
    gpu_capabilities: Optional[List[List[str]]] = None,
) -> ContainerRun:
    """Run a container in local.

    Args:
        image_uri (str):  A docker image uri.
        container_name (str, optional): Name of the container.
        port (int, optional): The port to expose.
        environment_variables (Dict[str, str], optional): Environment variables to set
            in the container.
        command (Union[List[str], str], optional): Command to run the container.
        entry_point (Union[List[str], str], optional): Entry point to run the container.
        volumes (Dict[str, Any], optional): Volumes to mount in the container.
        working_dir (str, optional): Working directory in the container.
        gpu_count (int, optional): Number of GPU devices to request. Set to -1 to
            request all available devices.
            To use GPU, set either ``gpu_count`` or ``gpu_device_ids``.
        gpu_device_ids (List[str], optional): List of strings for GPU device IDs,
            corresponding to `NVIDIA_VISIBLE_DEVICES` in the NVIDIA Runtime.
            To use GPU, set either ``gpu_count`` or ``gpu_device_ids``.
        gpu_capabilities (List[List[str]], optional): This parameter corresponds to
            `NVIDIA_DRIVER_CAPABILITIES` in the NVIDIA Runtime. The default value is
             ``[["compute", "utility"]]`` if ``gpu_device_ids`` or ``gpu_count`` is set.
             Available capabilities for the NVIDIA driver can be found in
            https://docs.nvidia.com/datacenter/cloud-native/container-toolkit/user-guide.html#driver-capabilities.

    Returns:
        ContainerRun: A ContainerRun object.

    """
    try:
        import docker
    except ImportError:
        raise ImportError("Please install docker first: pip install docker")

    client = docker.from_env()
    # use a random host port.
    host_port = randint(49152, 65535)

    if gpu_count or gpu_device_ids or gpu_capabilities:
        if not gpu_capabilities:
            gpu_capabilities = [["compute", "utility"]]
        device_requests = [
            docker.types.DeviceRequest(
                count=gpu_count,
                device_ids=gpu_device_ids,
                capabilities=gpu_capabilities,
            )
        ]
    else:
        device_requests = []

    container = client.containers.run(
        name=container_name,
        entrypoint=entry_point,
        image=image_uri,
        command=command,
        environment=environment_variables,
        ports={port: host_port} if port else None,
        volumes=volumes,
        working_dir=working_dir,
        detach=True,
        device_requests=device_requests,
    )
    container_run = ContainerRun(
        container=container,
        port=host_port,
    )
    return container_run



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import webbrowser
from abc import ABCMeta, abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from .common import git_utils
from .common.consts import FileSystemInputScheme, JobType
from .common.logging import get_logger
from .common.utils import is_local_run_instance_type, make_list_resource_iterator
from .job import (
    AlgorithmSpec,
    Channel,
    HyperParameterDefinition,
    LocalTrainingJob,
    TrainingJob,
    UriOutput,
    _TrainingJobSubmitter,
)
from .job._training_job import (
    DEFAULT_CHECKPOINT_CHANNEL_NAME,
    DEFAULT_OUTPUT_MODEL_CHANNEL_NAME,
    DEFAULT_TENSORBOARD_CHANNEL_NAME,
    ExperimentConfig,
    ResourceType,
    SpotSpec,
    UserVpcConfig,
)
from .model import InferenceSpec, Model, ResourceConfig
from .predictor import Predictor
from .serializers import SerializerBase
from .session import Session, get_default_session

logger = get_logger(__name__)


class HyperParameterType(object):
    """Hyperparameter type."""

    INT = "Int"
    FLOAT = "Float"
    STRING = "String"
    BOOL = "Boolean"
    CHANNEL = "Channel"

    @classmethod
    def convert(
        cls,
        hp_value: Any,
        hp_type: str,
    ):
        """Convert hyperparameter value to the specified type."""
        if hp_type == cls.INT:
            hp_value = int(hp_value)
        elif hp_type == cls.FLOAT:
            hp_value = float(hp_value)
        return hp_value


class FileSystemInputBase(metaclass=ABCMeta):
    """Base class for FileSystemInput."""

    @abstractmethod
    def to_input_uri(self):
        pass


class FileSystemInput(FileSystemInputBase):
    """FileSystemInput is used to mount a Standard/Extreme NAS file system for a
    TrainingJob.

    Examples::

        est = Estimator(
            image_uri="<TrainingImageUri>",
            command="sh train.sh",
            instance_type="ecs.c6.xlarge",
        )

        est.fit({
            "input": FileSystemInput(
                file_system_id="<FileSystemId>",
                directory_path="/path/to/data/"),
        })

    """

    def __init__(self, file_system_id: str, directory_path: Optional[str] = None):
        self.file_system_id = file_system_id
        self.directory_path = (
            directory_path.lstrip("/")
            if directory_path and directory_path.startswith("/")
            else directory_path
        )

    def to_input_uri(self):
        """Convert FileSystemInput to input uri used for TrainingJob."""
        region_id = get_default_session().region_id
        return "{schema}://{file_system_id}.{region_id}/{directory}".format(
            schema=FileSystemInputScheme.NAS,
            file_system_id=self.file_system_id,
            region_id=region_id,
            directory=self.directory_path or "",
        )


class CpfsFileSystemInput(FileSystemInputBase):
    """CpfsFileSystemInput is used to mount a CPFS file system for a TrainingJob.

    More details about CPFS, please refer to documentation:
    https://help.aliyun.com/product/111536.html

    Examples::

        est = Estimator(
            image_uri="<TrainingImageUri>",
            command="sh train.sh",
            instance_type="ecs.c6.xlarge",
        )

        est.fit(
            inputs={"train": CpfsFileSystemInput(
                file_system_id="<FileSystemId>",
                protocol_service_id="<ProtocolServiceId>",
                export_id="<ExportId>",
            )},
        )

    """

    def __init__(self, file_system_id: str, protocol_service_id: str, export_id: str):
        """Initialize CpfsFileSystemInput.

        Args:
            file_system_id (str): CPFS file system id.
            protocol_service_id (str): CPFS protocol service id.
            export_id (str): CPFS export id.
        """

        self.file_system_id = file_system_id
        self.protocol_service_id = protocol_service_id
        self.export_id = export_id

    def to_input_uri(self):
        """Convert CpfsFileSystemInput instance to input uri used for TrainingJob."""
        region_id = get_default_session().region_id
        return (
            "{schema}://{file_system_id}.{region_id}/{protocol_service_id}/"
            "{export_id}/".format(
                schema=FileSystemInputScheme.CPFS,
                file_system_id=self.file_system_id,
                region_id=region_id,
                protocol_service_id=self.protocol_service_id,
                export_id=self.export_id,
            )
        )


class EstimatorBase(_TrainingJobSubmitter, metaclass=ABCMeta):
    """EstimatorBase is the base class for other Estimator classes, such as Estimator.

    The EstimatorBase class contains common attributes and methods for all estimators,
    such as the hyperparameters, instance type, instance count, etc. The EstimatorBase
    class is not intended to be used directly, please use the other Estimator classes
    instead.

    """

    def __init__(
        self,
        hyperparameters: Optional[Dict[str, Any]] = None,
        base_job_name: Optional[str] = None,
        max_run_time: Optional[int] = None,
        output_path: Optional[str] = None,
        checkpoints_path: Optional[str] = None,
        environments: Optional[Dict[str, str]] = None,
        requirements: Optional[List[str]] = None,
        instance_type: Optional[str] = None,
        spot_spec: Optional[SpotSpec] = None,
        instance_spec: Optional[Dict] = None,
        resource_id: Optional[Dict] = None,
        resource_type: Optional[Union[str, ResourceType]] = None,
        instance_count: Optional[int] = None,
        user_vpc_config: Optional[UserVpcConfig] = None,
        experiment_config: Optional[ExperimentConfig] = None,
        settings: Optional[Dict[str, Any]] = None,
        labels: Optional[Dict[str, str]] = None,
        session: Optional[Session] = None,
    ):
        """EstimatorBase constructor.

        Args:
            hyperparameters (dict, optional): A dictionary that represents the
                hyperparameters used in the training job. The hyperparameters will be
                stored in the `/ml/input/config/hyperparameters.json` as a JSON
                dictionary in the training container.
            base_job_name (str, optional): The base name used to generate the training
                job name.
            max_run_time (int, optional): The maximum time in seconds that the training
                job can run. The training job will be terminated after the time is
                reached (Default None).
            output_path (str, optional): An OSS URI to store the outputs of the training
                jobs. If not provided, an OSS URI will be generated using the default
                OSS bucket in the session. When the `estimator.fit` method is called,
                a specific OSS URI under the output_path for each channel is generated
                and mounted to the training container.

                A completed training container directory structure example::

                    /ml
                    |-- usercode            			// User source code directory.
                    |   |-- requirements.txt
                    |   `-- train.py
                    |-- input               			// TrainingJob input
                    |   `-- config
                    |       |-- hyperparameters.json	// Hyperparameters in JSON
                    |       |                           // dictionary format for the
                    |       |                           // TrainingJob
                    |       |
                    |   `-- data            			// TrainingJob input channels
                    |       |                           // `/ml/input/data/` is a input
                    |       |                           // channel, and the directory
                    |       |                           // name is the channel name.
                    |       |                           // Each directory under the
                    |       |-- test-data
                    |       |   `-- test.csv
                    |       `-- train-data
                    |           `-- train.csv
                    `-- output              			// TrainingJob output channels.
                            |                           // Each directory under the
                            |                           // `/ml/output/` is an output
                            |                           // channel, and the directory
                            |                           // name is the channel name.
                            `-- model
                            `-- checkpoints
            checkpoints_path (str, optional): An OSS URI that stores the checkpoint of the
                training job. If provided, the OSS URI will be mounted to the directory
                `/ml/output/checkpoints/`.
            environments: A dictionary that maps environment variable names to their values.
                This optional field allows you to provide a set of environment variables that will be
                applied to the context where the code is executed.
            requirements (list, optional): An optional list of strings that specifies the Python
                package dependencies with their versions. Each string in the list should be in the format
                'package' or 'package==version'. This is similar to the contents of a requirements.txt file used
                in Python projects. If requirements.txt is provided in user code directory, requirements
                will override the conflict dependencies directly.
            resource_type (str, optional): The resource type used to run the training job.
                By default, general computing resource is used. If the resource_type is
                'Lingjun', Lingjun computing resource is used.
            instance_type (str, optional): The machine instance type used to run the
                training job. To view the supported machine instance types, please refer
                to the document:
                https://help.aliyun.com/document_detail/171758.htm#section-55y-4tq-84y.
                If the instance_type is "local", the training job is executed locally
                using docker.
            spot_spec (:class:`pai.job.SpotSpec`, optional): The specification of the spot
                instance used to run the training job. If provided, the training job will
                use the spot instance to run the training job.
            instance_count (int): The number of machines used to run the training job.
            user_vpc_config (:class:`pai.estimator.UserVpcConfig`, optional): The VPC
                configuration used to enable the training job instance to connect to the
                specified user VPC. If provided, an Elastic Network Interface (ENI) will
                be created and attached to the training job instance, allowing the
                instance to access the resources within the specified VPC. Default to
                None.
            experiment_config (:class:`pai.estimator.ExperimentConfig`, optional): The
                experiment configuration used to construct the relationship between the
                training job and the experiment. If provided, the training job will belong
                to the specified experiment, in which case the training job will use
                artifact_uri of experiment as default output path. Default to None.
            settings (dict, optional): A dictionary that represents the additional settings
                for job, such as AIMaster configurations.
            labels (Dict[str, str], optional): A dictionary that maps label names to
                their values. This optional field allows you to provide a set of labels
                that will be applied to the training job.
            session (Session, optional): A PAI session instance used for communicating
                with PAI service.

        """
        self.hyperparameters = hyperparameters or dict()
        self.checkpoints_path = checkpoints_path
        self.session = session or get_default_session()
        super().__init__(
            base_job_name=base_job_name,
            output_path=output_path,
            experiment_config=experiment_config,
            instance_type=instance_type,
            instance_count=instance_count,
            resource_id=resource_id,
            resource_type=resource_type,
            spot_spec=spot_spec,
            instance_spec=instance_spec,
            user_vpc_config=user_vpc_config,
            max_run_time=max_run_time,
            environments=environments,
            requirements=requirements,
            settings=settings,
            labels=labels,
        )

    def set_hyperparameters(self, **kwargs):
        """Set hyperparameters for the training job.

        Args:
            **kwargs: Hyperparameters in key-value pairs.
        """
        self.hyperparameters.update(**kwargs)

    def _gen_job_display_name(self, job_name=None):
        """Generate job display name."""
        if job_name:
            return job_name
        ts = datetime.now().strftime("%Y%m%d_%H%M%S")
        return "{}_{}".format(self.base_job_name or "training_job", ts)

    @abstractmethod
    def fit(
        self, inputs: Dict[str, Any] = None, wait: bool = True, show_logs: bool = True
    ):
        """Submit a training job with the given input data."""

    def model_data(self) -> str:
        """Model data output path.

        Returns:
            str: A string in OSS URI format refers to the output model of the submitted
                job.
        """
        if not self.latest_job:
            raise RuntimeError(
                "No TrainingJob for the estimator, output model data not found."
            )

        if not self.latest_job.is_succeeded():
            logger.warning(
                "The TrainingJob is currently not in a succeeded status, which means"
                " that the model data output may not be accessible."
            )

        return self.latest_job.output_path(
            channel_name=DEFAULT_OUTPUT_MODEL_CHANNEL_NAME
        )

    def checkpoints_data(self) -> str:
        """Checkpoints data output path.

        Returns:
            str: A string in OSS URI format refers to the checkpoints of submitted
                training job.
        """
        if not self.latest_job:
            raise RuntimeError(
                "No TrainingJob for the Estimator, output checkpoints data path "
                "not found."
            )
        return self.latest_job.output_path(channel_name=DEFAULT_CHECKPOINT_CHANNEL_NAME)

    def tensorboard_data(self) -> str:
        """Output TensorBoard logs path.

        Returns:
            str: A string in OSS URI format refers to the tensorboard log of submitted
                training job.
        """
        if not self.latest_job:
            raise RuntimeError(
                "No TrainingJob for the Estimator, output TensorBoard logs data path"
                " not found."
            )
        return self.latest_job.output_path(
            channel_name=DEFAULT_TENSORBOARD_CHANNEL_NAME,
        )

    def tensorboard(self, wait=True):
        """Launch a TensorBoard Application to view the output TensorBoard logs.

        Args:
            wait (bool): Specifies whether to block until the TensorBoard is running.

        Returns:
            :class:`pai.tensorboard.TensorBoard`: A TensorBoard instance.
        """
        from pai.tensorboard import TensorBoard

        if not self.latest_job:
            raise RuntimeError("Could not find a submitted training job.")

        source_type = "TrainingJob"
        if isinstance(self.latest_job, LocalTrainingJob):
            raise RuntimeError("Local training job does not support tensorboard.")
        res = self.session.tensorboard_api.list(
            source_type=source_type,
            source_id=self.latest_job.training_job_id,
        )

        if res.items:
            if len(res.items) > 1:
                logger.warning(
                    "Found multiple TensorBoard instances for the submitted training "
                    "job, use the first one."
                )
            tb_id = res.items[0]["TensorboardId"]
            tb = TensorBoard(tensorboard_id=tb_id, session=self.session)
            tb.start(wait=wait)
        else:
            tb = TensorBoard.create(
                uri=self.tensorboard_data(),
                wait=wait,
                display_name=self.latest_job.training_job_name,
                source_id=self.latest_job.training_job_id,
                source_type=source_type,
                session=self.session,
            )

        # Open the TensorBoard in the default browser.
        webbrowser.open(tb.app_uri)
        return tb

    def create_model(self, inference_spec: Union[InferenceSpec, Dict]) -> Model:
        """Create a Model object using output model of the training job.

        Args:
            inference_spec (InferenceSpec): A ``InferenceSpec`` instance that describe
             how to create a prediction service with the output model.

        Returns:
            :class:`pai.model.Model`: A ``Model`` object.
        """

        if isinstance(inference_spec, Dict):
            inference_spec = InferenceSpec.from_dict(inference_spec)

        m = Model(
            model_data=self.model_data(),
            inference_spec=inference_spec,
            session=self.session,
        )
        return m

    def deploy(
        self,
        service_name: str,
        inference_spec: InferenceSpec,
        instance_type: Optional[str] = None,
        instance_count: int = 1,
        resource_config: Optional[Union[ResourceConfig, Dict[str, int]]] = None,
        resource_id: str = None,
        options: Optional[Dict[str, Any]] = None,
        serializer: SerializerBase = None,
        wait=True,
    ) -> Predictor:
        """Deploy the output model to create an online prediction service.

        Args:
            service_name (str): Name for the online prediction service.
            inference_spec (InferenceSpec): A ``InferenceSpec`` instance used for
                creating the service.
            instance_type (str, optional): The machine instance type for the service.
            instance_count (int): Number of machine instance count.
            resource_config (Union[ResourceConfig, Dict[str, int]], optional): Resource
                config for each prediction service instance.
            resource_id (str, optional): The ID of the resource group. If not provided,
                the prediction service is deployed to ``public resource group``.
            serializer (SerializerBase): A SerializerBase instance used to serialize
                the prediction request data and deserialize the response data.
            options (Dict[str, Any], optional): Additional options for the prediction
                service.
            wait (bool): If true, wait until the service is ready (Default True).

        Returns:
            :class:`pai.predictor.Predictor`: A predictor instance refers to the created
                prediction service.
        """
        m = Model(
            model_data=self.model_data(),
            inference_spec=inference_spec,
        )

        p = m.deploy(
            service_name=service_name,
            instance_type=instance_type,
            instance_count=instance_count,
            resource_config=resource_config,
            resource_id=resource_id,
            serializer=serializer,
            options=options,
            wait=wait,
        )
        return p


class Estimator(EstimatorBase):
    """The Estimator object is responsible for submitting TrainingJob.

    The Estimator helps to run a training script in the PAI Training Service with a
    specific image.

    Example::

        est = Estimator(
            source_dir="./train/src/",
            command="python train.py",
            image_uri = training_image_uri,
            instance_type="ecs.c6.xlarge",
            hyperparameters={
                "n_estimators": 50,
                "objective": "binary:logistic",
                "max_depth": 5,
                "eval_metric": "auc",
            },
            output_path="oss://{YOUR_BUCKET_NAME}/pai/training_job/output_path",
        )

        est.fit(inputs={
            "train": "oss://{YOUR_BUCKET_NAME}/path/to/train-data",
            "test": "oss://{YOUR_BUCKET_NAME}/path/to/test-data",
        })

        print(est.model_data())

    """

    def __init__(
        self,
        image_uri: str,
        command: Union[str, List[str]],
        source_dir: Optional[str] = None,
        git_config: Optional[Dict[str, str]] = None,
        job_type: str = JobType.PyTorchJob,
        hyperparameters: Optional[Dict[str, Any]] = None,
        environments: Optional[Dict[str, str]] = None,
        requirements: Optional[List[str]] = None,
        base_job_name: Optional[str] = None,
        max_run_time: Optional[int] = None,
        checkpoints_path: Optional[str] = None,
        output_path: Optional[str] = None,
        metric_definitions: Optional[List[Dict[str, str]]] = None,
        instance_type: Optional[str] = None,
        instance_count: Optional[int] = None,
        user_vpc_config: Optional[UserVpcConfig] = None,
        experiment_config: Optional[ExperimentConfig] = None,
        resource_id: Optional[str] = None,
        session: Optional[Session] = None,
        **kwargs,
    ):
        """Estimator constructor.

        Args:
            image_uri (str): The image used in the training job. It can be an image
                provided by PAI or a user customized image. To view the images provided
                by PAI, please refer to the document:
                https://help.aliyun.com/document_detail/202834.htm.
            command (Union[str, List[str]]): The command used to run the training job.
            source_dir (str, optional): The local source code directory used in the
                training job. The directory will be packaged and uploaded to an OSS
                bucket, then downloaded to the `/ml/usercode` directory in the training
                job container. If there is a `requirements.txt` file in the source code
                directory, the corresponding dependencies will be installed before the
                training script runs.

                If 'git_config' is provided, 'source_dir' should be a relative location
                to a directory in the Git repo. With the following GitHub repo directory
                structure:

                .. code::

                    |----- README.md
                    |----- src
                             |----- train.py
                             |----- test.py

                if you need 'src' directory as the source code directory, you can assign
                source_dir='./src/'.
            git_config (Dict[str, str]): Git configuration used to clone the repo.
                Including ``repo``, ``branch``, ``commit``, ``username``, ``password``
                and ``token``. The ``repo`` is required. All other fields are optional.
                ``repo`` specifies the Git repository. If you don't provide ``branch``,
                the default value 'master' is used. If you don't provide ``commit``, the
                latest commit in the specified branch is used. ``username``, ``password``
                and ``token`` are for authentication purpose.
                For example, the following config:

                .. code:: python

                    git_config = {
                        'repo': 'https://github.com/modelscope/modelscope.git',
                        'branch': 'master',
                        'commit': '9bfc4a9d83c4beaf8378d0a186261ffc1cd9f960'
                    }

                results in cloning the git repo specified in 'repo', then checking out
                the 'master' branch, and checking out the specified commit.
            job_type (str): The type of job, which can be TFJob, PyTorchJob, XGBoostJob,
                etc.
            hyperparameters (dict, optional): A dictionary that represents the
                hyperparameters used in the training job. The hyperparameters will be
                stored in the `/ml/input/config/hyperparameters.json` as a JSON
                dictionary in the training container.
            environments: A dictionary that maps environment variable names to their values.
                This optional field allows you to provide a set of environment variables that will be
                applied to the context where the code is executed.
            requirements (list, optional): An optional list of strings that specifies the Python
                package dependencies with their versions. Each string in the list should be in the format
                'package' or 'package==version'. This is similar to the contents of a requirements.txt file used
                in Python projects. If requirements.txt is provided in user code directory, requirements
                will override the conflict dependencies directly.
            instance_type (str): The machine instance type used to run the training job.
                To view the supported machine instance types, please refer to the
                document:
                https://help.aliyun.com/document_detail/171758.htm#section-55y-4tq-84y.
                If the instance_type is "local", the training job is executed locally
                using docker.
            max_run_time (int, optional): The maximum time in seconds that the training
                job can run. The training job will be terminated after the time is
                reached (Default None).
            instance_count (int): The number of machines used to run the training job.
            base_job_name (str, optional): The base name used to generate the training
                job name.
            checkpoints_path (str, optional): An OSS URI that stores the checkpoint of the
                training job. If provided, the OSS URI will be mounted to the directory
                `/ml/output/checkpoints/`.
            user_vpc_config (:class:`pai.estimator.UserVpcConfig`, optional): The VPC
                configuration used to enable the training job instance to connect to the
                specified user VPC. If provided, an Elastic Network Interface (ENI) will
                be created and attached to the training job instance, allowing the
                instance to access the resources within the specified VPC. Default to
                None.
            experiment_config(:class:`pai.estimator.ExperimentConfig`, optional): The
                experiment configuration used to construct the relationship between the
                training job and the experiment. If provided, the training job will belong
                to the specified experiment, in which case the training job will use
                artifact_uri of experiment as default output path. Default to None.
            output_path (str, optional): An OSS URI to store the outputs of the training
                jobs. If not provided, an OSS URI will be generated using the default
                OSS bucket in the session. When the `estimator.fit` method is called,
                a specific OSS URI under the output_path for each channel is generated
                and mounted to the training container.

                A completed training container directory structure example::

                    /ml
                    |-- usercode            			// User source code directory.
                    |   |-- requirements.txt
                    |   `-- train.py
                    |-- input               			// TrainingJob input
                    |   `-- config
                    |       |-- hyperparameters.json	// Hyperparameters in JSON
                    |       |                           // dictionary format for the
                    |       |                           // TrainingJob
                    |       |
                    |   `-- data            			// TrainingJob input channels
                    |       |                           // `/ml/input/data/` is a input
                    |       |                           // channel, and the directory
                    |       |                           // name is the channel name.
                    |       |                           // Each directory under the
                    |       |-- test-data
                    |       |   `-- test.csv
                    |       `-- train-data
                    |           `-- train.csv
                    `-- output              			// TrainingJob output channels.
                            |                           // Each directory under the
                            |                           // `/ml/output/` is an output
                            |                           // channel, and the directory
                            |                           // name is the channel name.
                            `-- model
                            `-- checkpoints

            metric_definitions (List[Dict[str, Any]): A list of dictionaries that
                defines the metrics used to evaluate the training jobs. Each dictionary
                contains two keys: "Name" for the name of the metric, and "Regex" for
                the regular expression used to extract the metric from the logs of the
                training job. The regular expression should contain only one capture
                group that is responsible for extracting the metric value.

                Example::

                    metric_definitions=[
                        {
                            "Name": "accuracy",
                            "Regex": r".*accuracy="
                                     r"([-+]?[0-9]*.?[0-9]+(?:[eE][-+]?[0-9]+)?).*",
                        },
                        {
                            "Name": "train-accuracy",
                            "Regex": r".*validation_0-auc="
                                     r"([-+]?[0-9]*.?[0-9]+(?:[eE][-+]?[0-9]+)?).*",
                        },
                    ]
            session (Session, optional): A PAI session instance used for communicating
                with PAI service.

        """
        self.image_uri = image_uri
        self.command = command
        self.source_dir = source_dir
        self.git_config = git_config
        self.job_type = job_type if job_type else JobType.PyTorchJob
        self.metric_definitions = metric_definitions

        super(Estimator, self).__init__(
            hyperparameters=hyperparameters,
            environments=environments,
            requirements=requirements,
            base_job_name=base_job_name,
            max_run_time=max_run_time,
            output_path=output_path,
            checkpoints_path=checkpoints_path,
            instance_type=instance_type,
            instance_count=instance_count,
            user_vpc_config=user_vpc_config,
            experiment_config=experiment_config,
            resource_id=resource_id,
            session=session,
            **kwargs,
        )

    def training_image_uri(self) -> str:
        """Return the Docker image to use for training.

        The fit() method, that does the model training, calls this method to
        find the image to use for model training.

        Returns:
            str: The URI of the Docker image.
        """
        return self.image_uri

    def _prepare_for_training(self):
        """Update args before starting the training job."""
        if self.git_config:
            updated_args = git_utils.git_clone_repo(
                git_config=self.git_config,
                source_dir=self.source_dir,
            )
            self.source_dir = updated_args["source_dir"]

    def _build_algorithm_spec(
        self, code_input, inputs: Dict[str, Any]
    ) -> AlgorithmSpec:
        """Build a temporary AlgorithmSpec used for submitting the TrainingJob."""
        algorithm_spec = AlgorithmSpec(
            command=(
                self.command
                if isinstance(self.command, list)
                else ["sh", "-c", self.command]
            ),
            image=self.training_image_uri(),
            job_type=self.job_type,
            metric_definitions=self.metric_definitions,
            code_dir=code_input,
            output_channels=self._default_training_output_channels(),
            input_channels=[
                Channel(name=channel_name, required=False)
                for channel_name in inputs.keys()
            ],
        )
        return algorithm_spec

    def fit(
        self,
        inputs: Dict[str, Any] = None,
        outputs: Dict[str, Any] = None,
        wait: bool = True,
        show_logs: bool = True,
        job_name: Optional[str] = None,
    ) -> Union[TrainingJob, LocalTrainingJob]:
        """Submit a training job with the given input data.

        Args:
            inputs (Dict[str, Any]): A dictionary representing the input data for the
                training job. Each key/value pair in the dictionary is an input channel,
                the key is the channel name, and the value is the input data. The input
                data can be an OSS URI or a NAS URI object and will be mounted to the
                `/ml/input/data/{channel_name}` directory in the training container.
             outputs (Dict[str, Any]): A dictionary representing the output locations for
                the training job. Each key/value pair in the dictionary is an output channel,
                the key is the channel name, and the value is the output data location.
            wait (bool): Specifies whether to block until the training job is completed,
                either succeeded, failed, or stopped. (Default True).
            show_logs (bool): Specifies whether to show the logs produced by the
                training job (Default True).
            job_name (str, optional): The name of the training job.

        Returns:
            :class:`pai.job.TrainingJob` or :class:`pai.job.LocalTrainingJob`: A
                submitted training job.

        Raises:
            UnExpectedStatusException: If the training job fails.

        """
        inputs = inputs or dict()
        self._prepare_for_training()
        job_name = self.job_name(job_name=job_name)
        if is_local_run_instance_type(self.instance_type):
            return self._local_run(
                job_name=job_name,
                inputs=inputs,
                instance_type=self.instance_type,
                wait=wait,
            )
        return self._fit(
            inputs=inputs,
            outputs=outputs,
            job_name=job_name,
            wait=wait,
            show_logs=show_logs,
        )

    def _fit(
        self,
        job_name,
        inputs: Dict[str, Any],
        outputs: Dict[str, Any],
        wait: bool = True,
        show_logs: bool = True,
    ) -> TrainingJob:
        # prepare input code.
        code_input = self._build_code_input(job_name, source_dir=self.source_dir)
        algo_spec = self._build_algorithm_spec(
            code_input=code_input,
            inputs=inputs,
        )
        inputs = self.build_inputs(
            inputs=inputs,
            input_channels=algo_spec.input_channels,
        )

        outputs = outputs or {}
        if self.checkpoints_path:
            outputs.update({DEFAULT_CHECKPOINT_CHANNEL_NAME: self.checkpoints_path})

        outputs = self.build_outputs(
            job_name=job_name,
            output_channels=algo_spec.output_channels,
            outputs=outputs,
        )

        return self._submit(
            job_name=job_name,
            algorithm_spec=algo_spec,
            instance_spec=self.instance_spec,
            instance_type=self.instance_type,
            instance_count=self.instance_count,
            resource_id=self.resource_id,
            hyperparameters=self.hyperparameters,
            environments=self.environments,
            requirements=self.requirements,
            max_run_time=self.max_run_time,
            inputs=inputs,
            outputs=outputs,
            user_vpc_config=self.user_vpc_config if self.user_vpc_config else None,
            experiment_config=(
                self.experiment_config if self.experiment_config else None
            ),
            labels=self.labels,
            wait=wait,
            show_logs=show_logs,
        )

    def _local_run(
        self,
        job_name,
        instance_type: str,
        inputs: Dict[str, Any] = None,
        wait: bool = True,
    ) -> "LocalTrainingJob":
        if self.instance_count > 1:
            raise RuntimeError("Local training job only supports single instance.")

        training_job = LocalTrainingJob(
            estimator=self,
            inputs=inputs,
            job_name=job_name,
            instance_type=instance_type,
        )
        training_job.run()
        if wait:
            training_job.wait()
        return training_job


class AlgorithmEstimator(EstimatorBase):
    """Handle training jobs with algorithms

    The AlgorithmEstimator provides a simple way for submitting training jobs with
    algorithms.

    Example::

        # Create an AlgorithmEstimator with built-in algorithms
        est = AlgorithmEstimator(
            algorithm_name="pai-algorithm-test",
            algorithm_version="0.1.0",
            algorithm_provider="pai",
        )

        # Inspect the definition of hyperparameters, input channels and output channels
        print(est.hyperparameter_definitions)
        print(est.input_channel_definitions)
        print(est.output_channel_definitions)
        print(est.supported_instance_types)

        # Submit a training job
        est.fit(
            inputs={
                "train": "oss://bucket/path/to/train/data",
                "test": "oss://bucket/path/to/test/data",
            },
        )

        # Inspect all outputs data
        print(est.get_outputs_data())

    """

    def __init__(
        self,
        algorithm_name: Optional[str] = None,
        algorithm_version: Optional[str] = None,
        algorithm_provider: Optional[str] = None,
        algorithm_spec: Optional[AlgorithmSpec] = None,
        hyperparameters: Optional[Dict[str, Any]] = None,
        environments: Optional[Dict[str, str]] = None,
        requirements: Optional[List[str]] = None,
        base_job_name: Optional[str] = None,
        max_run_time: Optional[int] = None,
        output_path: Optional[str] = None,
        instance_type: Optional[str] = None,
        instance_count: Optional[int] = None,
        user_vpc_config: Optional[UserVpcConfig] = None,
        session: Optional[Session] = None,
        instance_spec: Optional[Dict[str, Union[int, str]]] = None,
        **kwargs,
    ):
        """Initialize an AlgorithmEstimator.

        Args:
            algorithm_name (str, optional): The name of the registered algorithm. If not
                provided, the algorithm_spec must be provided.
            algorithm_version (str, optional): The version of the algorithm. If not
                provided, the latest version of the algorithm will be used. If algorithm
                name is not provided, this argument will be ignored.
            algorithm_provider (str, optional): The provider of the algorithm.
                Currently, only "pai" or None are supported. Set it to "pai" to retrieve
                 a PAI official algorithm. If not provided, the default provider is
                 user's PAI account. If algorithm name is not provided, this argument
                  will be ignored.
            algorithm_spec (AlgorithmSpec, optional): A temporary algorithm spec.
                Required if algorithm_name is not provided.
            hyperparameters (dict, optional): A dictionary that represents the
                hyperparameters used in the training job. Default hyperparameters will
                be retrieved from the algorithm definition.
            environments: A dictionary that maps environment variable names to their values.
                This optional field allows you to provide a set of environment variables that will be
                applied to the context where the code is executed.
            requirements (list, optional): An optional list of strings that specifies the Python
                package dependencies with their versions. Each string in the list should be in the format
                'package' or 'package==version'. This is similar to the contents of a requirements.txt file used
                in Python projects. If requirements.txt is provided in user code directory, requirements
                will override the conflict dependencies directly.
            base_job_name (str, optional): The base name used to generate the training
                job name. If not provided, a default job name will be generated.
            max_run_time (int, optional): The maximum time in seconds that the training
                job can run. The training job will be terminated after the time is
                reached (Default None).
            output_path (str, optional): An OSS URI to store the outputs of the training
                jobs. If not provided, an OSS URI will be generated using the default
                OSS bucket in the session. When the `estimator.fit` method is called,
                a specific OSS URI under the output_path for each channel is generated
                and mounted to the training container.
            instance_type (str, optional): The machine instance type used to run the
                training job. If not provider, the default instance type will be
                retrieved from the algorithm definition. To view the supported machine
                instance types, please refer to the document:
                https://help.aliyun.com/document_detail/171758.htm#section-55y-4tq-84y.
            instance_count (int, optional): The number of machines used to run the
                training job. If not provider, the default instance count will be
                retrieved from the algorithm definition.
            user_vpc_config (:class:`pai.estimator.UserVpcConfig`, optional): The VPC
                configuration used to enable the training job instance to connect to the
                specified user VPC. If provided, an Elastic Network Interface (ENI) will
                be created and attached to the training job instance, allowing the
                instance to access the resources within the specified VPC. Default to
                None.
            session (:class:`pai.session.Session`, optional): A PAI session object
                used for interacting with PAI Service.
        """
        self._check_args(
            algorithm_name=algorithm_name,
            algorithm_spec=algorithm_spec,
        )

        self.session = session or get_default_session()

        # Use _algo_spec to store the algorithm spec for inner use no matter the
        # algorithm_name is provided or the algorithm_spec is provided.
        # If algorithm_name is provided, retrieve the algorithm spec from the registry.
        if algorithm_name:
            _algo_version = self._get_algo_version(
                algorithm_name=algorithm_name,
                algorithm_version=algorithm_version,
                algorithm_provider=algorithm_provider,
            )
            self._algo_spec = AlgorithmSpec.model_validate(
                _algo_version["AlgorithmSpec"]
            )
            self.algorithm_name = _algo_version["AlgorithmName"]
            self.algorithm_version = _algo_version["AlgorithmVersion"]
            self.algorithm_provider = _algo_version["AlgorithmProvider"]
            self.algorithm_spec = None
        # If algorithm_name is not provided, use the provided algorithm_spec.
        else:
            self._algo_spec = algorithm_spec
            self.algorithm_name = None
            self.algorithm_version = None
            self.algorithm_provider = None
            self.algorithm_spec = algorithm_spec

        if not instance_type and not instance_spec:
            instance_type = self._get_default_training_instance_type()
        super(AlgorithmEstimator, self).__init__(
            hyperparameters=self._get_hyperparameters(hyperparameters),
            environments=environments,
            requirements=requirements,
            base_job_name=base_job_name,
            max_run_time=max_run_time,
            output_path=output_path,
            instance_type=instance_type,
            instance_count=instance_count,
            session=session,
            user_vpc_config=user_vpc_config,
            instance_spec=instance_spec,
            **kwargs,
        )

    # TODO: check if the hyperparameters are valid
    def set_hyperparameters(self, **kwargs):
        """Set hyperparameters for the algorithm training."""
        super(AlgorithmEstimator, self).set_hyperparameters(**kwargs)

    @property
    def hyperparameter_definitions(self) -> List[HyperParameterDefinition]:
        """Get the hyperparameter definitions from the algorithm spec."""
        res = self._algo_spec.hyperparameter_definitions
        return res

    @property
    def input_channel_definitions(self) -> List[Channel]:
        """Get the input channel definitions from the algorithm spec."""
        res = self._algo_spec.input_channels
        return res

    @property
    def output_channel_definitions(self) -> List[Channel]:
        """Get the output channel definitions from the algorithm spec."""
        res = self._algo_spec.output_channels
        return res

    @property
    def supported_instance_types(self) -> List[str]:
        """Get the supported instance types from the algorithm spec."""
        return self._algo_spec.supported_instance_types

    def _check_args(
        self,
        algorithm_name: str,
        algorithm_spec: Optional[AlgorithmSpec],
    ):
        """Check the algorithm_name and algorithm_spec.

        If neither algorithm_name nor algorithm_spec is provided, raise a ValueError.
        If both algorithm_name and algorithm_spec are provided, use the algorithm_name
        by default and ignore the algorithm_spec.

        Args:
            algorithm_name (str): The name of the algorithm.
            algorithm_spec (AlgorithmSpec): The algorithm spec.
        """
        if not algorithm_name and not algorithm_spec:
            raise ValueError(
                "Either algorithm_name or algorithm_spec should be provided."
            )
        if algorithm_name and algorithm_spec:
            logger.warning(
                "Both a tuple of (algorithm_name, algorithm_version,"
                " algorithm_provider) and algorithm_spec are provided. Use the tuple of"
                " (algorithm_name, algorithm_version, algorithm_provider) by default."
                " The provided algorithm_spec will be ignored."
            )

    def _get_algo_version(
        self,
        algorithm_name: str,
        algorithm_version: Optional[str] = None,
        algorithm_provider: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Get the algorithm version object.

        Args:
            algorithm_name (str): The name of the algorithm.
            algorithm_version (str, optional): The version of the algorithm.
            algorithm_provider (str, optional): The provider of the algorithm.

        Returns:
            dict: A dict that represents algorithm version object.
        """
        if not algorithm_name:
            raise ValueError(
                "Parameter algorithm_name cannot be None or empty. Please provide a"
                " valid algorithm name."
            )

        resp_list_algo = self.session.algorithm_api.list(
            algorithm_name=algorithm_name,
            algorithm_provider=algorithm_provider,
        )
        if resp_list_algo.total_count == 0:
            raise ValueError(
                f"Could not find any algorithm with the specific"
                f" name='{algorithm_name}' and provider='{algorithm_provider}'."
                f" Please check the arguments."
            )
        algo_obj = resp_list_algo.items[0]
        algorithm_id = algo_obj["AlgorithmId"]

        if not algorithm_version:
            resp_list_algo_versions = self.session.algorithm_api.list_versions(
                algorithm_id=algorithm_id,
            )
            algo_version_obj = resp_list_algo_versions.items[-1]
            algorithm_version = algo_version_obj["AlgorithmVersion"]
            logger.warning(
                f"Parameter algorithm_version is not provided, the latest"
                f" version='{algorithm_version}' of the algorithm will be used."
            )

        resp_algo_version = self.session.algorithm_api.get_version(
            algorithm_id=algorithm_id,
            algorithm_version=algorithm_version,
        )
        return resp_algo_version

    def _get_hyperparameters(
        self, hyperparameters: Optional[Dict[str, Any]]
    ) -> Dict[str, Any]:
        """Get hyperparameters.

        Get the default hyperparameters from the algorithm spec and update it with the
        user provided hyperparameters.

        Args:
            hyperparameters (dict, optional): The user provided hyperparameters.

        Returns:
            dict: The dict of hyperparameters.
        """
        res = {}
        hps_def = self.hyperparameter_definitions
        if hps_def:
            # Get default hyperparameters.
            for hp in hps_def:
                hp_name = hp.name
                hp_value = hp.default_value
                hp_type = hp.type or "String"
                # For hyperparameters with type INT or FLOAT, if the default value is
                # empty, skip it.
                if (
                    hp_type in [HyperParameterType.INT, HyperParameterType.FLOAT]
                    and not hp_value
                ):
                    continue
                else:
                    # Convert the value to the corresponding type.
                    hp_value = HyperParameterType.convert(hp_value, hp_type)
                    res.update({hp_name: hp_value})
            # Update with user provided hyperparameters.
            for hp in hyperparameters if hyperparameters else {}:
                if hp not in res:
                    logger.warning(
                        f"Hyperparameter='{hp}' is not defined in hyperparameters"
                        f" definition. Make sure you are using the right"
                        f" hyperparameters and check the hyperparameter_definitions."
                    )
                    res.update({hp: hyperparameters[hp]})
                else:
                    res[hp] = hyperparameters[hp]
        else:
            # Use user provided hyperparameters.
            res = hyperparameters
        return res

    def _get_default_training_instance_type(self) -> str:
        """Get the default training instance type from the algorithm spec."""
        instance_generator = make_list_resource_iterator(
            self.session.job_api.list_ecs_specs
        )
        sup_instance_types = self.supported_instance_types
        machine_spec = next(
            (
                item
                for item in instance_generator
                if not sup_instance_types or item["InstanceType"] in sup_instance_types
            ),
            None,
        )
        if not machine_spec:
            raise RuntimeError(
                "No supported training instance type found. Please check the supported"
                " instance types."
            )
        return machine_spec["InstanceType"]

    def fit(
        self,
        inputs: Dict[str, Any] = None,
        outputs: Dict[str, Any] = None,
        wait: bool = True,
        show_logs: bool = True,
        job_name: Optional[str] = None,
    ) -> TrainingJob:
        """Submit a training job with the given input data.

        Args:
            inputs (Dict[str, Any]): A dictionary representing the input data for the
                training job. Each key/value pair in the dictionary is an input channel,
                the key is the channel name, and the value is the input data. The input
                data can be an OSS URI or a NAS URI object and will be mounted to the
                `/ml/input/data/{channel_name}` directory in the training container.
            wait (bool): Specifies whether to block until the training job is completed,
                either succeeded, failed, or stopped. (Default True).
            show_logs (bool): Whether to show the logs of the training job. Default to True.
                Note that the logs will be shown only when the `wait` is set to True.
            job_name (str, optional): The name of the training job.

        Returns:
            :class:`pai.training_job.TrainingJob`: The submitted training job.

        Raises:
            UnExpectedStatusException: If the training job fails.

        """
        job_name = self.job_name(job_name=job_name)
        input_configs = self.build_inputs(
            inputs,
            input_channels=self._algo_spec.input_channels,
        )
        output_configs = self.build_outputs(
            job_name,
            output_channels=self._algo_spec.output_channels,
            outputs=outputs,
        )
        return self._submit(
            instance_count=self.instance_count,
            instance_type=self.instance_type,
            instance_spec=self.instance_spec,
            resource_id=self.resource_id,
            job_name=job_name,
            hyperparameters=self.hyperparameters,
            max_run_time=self.max_run_time,
            inputs=input_configs,
            outputs=output_configs,
            environments=self.environments,
            requirements=self.requirements,
            algorithm_name=self.algorithm_name,
            algorithm_version=self.algorithm_version,
            algorithm_provider=self.algorithm_provider,
            algorithm_spec=self.algorithm_spec,
            user_vpc_config=(
                self.user_vpc_config.to_dict() if self.user_vpc_config else None
            ),
            experiment_config=(
                self.experiment_config.to_dict() if self.experiment_config else None
            ),
            labels=self.labels,
            wait=wait,
            show_logs=show_logs,
        )

    def get_outputs_data(self) -> Dict[str, str]:
        """Show all outputs data paths.

        Returns:
            dict[str, str]: A dictionary of all outputs data paths.
        """
        if not self.latest_job:
            raise RuntimeError(
                "Could not find a submitted training job. Please submit a training job"
                " before calling this method."
            )

        uri_outputs = [
            output
            for output in self.latest_job.outputs
            if isinstance(output, UriOutput)
        ]
        extra_outputs = [
            output
            for output in self.latest_job.outputs
            if not isinstance(output, UriOutput)
        ]

        if extra_outputs:
            logger.warning(
                "Extra outputs are provided in the training job, but only URI outputs"
                " are supported. The extra outputs will be ignored: %s",
                extra_outputs,
            )
        return {ch.name: ch.output_uri for ch in uri_outputs}



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import yaml


class NoAliasDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True


def dump(data):
    """Serialize input Python object with NoAliasDumper.

    Args:
        data: a Python object.

    Returns:
        str: Return the produced string.

    """
    return yaml.dump(data, Dumper=NoAliasDumper)


def dump_all(seq):
    """Serialize a sequence of Python objects with NoAliasDumper.

    Args:
        seq : A sequence of Python objects.

    Returns:
        str: Return the produced string.

    """
    return yaml.dump_all(seq, Dumper=NoAliasDumper)


def safe_load(stream):
    """Parse the first YAML document in the stream."""
    return yaml.safe_load(stream)



#  Copyright 2024 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ._local_training_job import LocalTrainingJob
from ._training_job import (
    AlgorithmSpec,
    Channel,
    CodeDir,
    ExperimentConfig,
    HyperParameterDefinition,
    InstanceSpec,
    ModelRecipeSpec,
    OssLocation,
    ResourceType,
    SpotSpec,
    SpotStrategy,
    TrainingJob,
    TrainingJobStatus,
    UriInput,
    UriOutput,
    UserVpcConfig,
    _TrainingJobSubmitter,
)

__all__ = [
    "TrainingJob",
    "ModelRecipeSpec",
    "TrainingJobStatus",
    "Channel",
    "HyperParameterDefinition",
    "OssLocation",
    "AlgorithmSpec",
    "CodeDir",
    "LocalTrainingJob",
    "UriOutput",
    "UserVpcConfig",
    "ExperimentConfig",
    "InstanceSpec",
    "UriInput",
    "SpotSpec",
    "ResourceType",
    "SpotStrategy",
]



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABCMeta
from typing import Any, Dict, List, Optional, Union

import six
from alibabacloud_tea_openapi.client import Client
from alibabacloud_tea_util.models import RuntimeOptions
from six import with_metaclass
from Tea.model import TeaModel

from ..common.logging import get_logger

logger = get_logger(__name__)


class ServiceName(object):
    # Service provided by PAI.
    PAI_DLC = "pai-dlc"
    PAI_EAS = "pai-eas"
    PAI_WORKSPACE = "aiworkspace"
    PAI_STUDIO = "pai"
    PAIFLOW = "paiflow"
    # Other services provided by Alibaba Cloud.
    STS = "sts"
    PAI_DSW = "pai-dsw"


class PAIRestResourceTypes(object):
    """Resource types provided by PAI REST API."""

    Dataset = "Dataset"
    DlcJob = "DlcJob"
    CodeSource = "CodeSource"
    Image = "Image"
    Service = "Service"
    Model = "Model"
    Workspace = "Workspace"
    Algorithm = "Algorithm"
    TrainingJob = "TrainingJob"
    Pipeline = "Pipeline"
    PipelineRun = "PipelineRun"
    TensorBoard = "TensorBoard"
    Experiment = "Experiment"


class ResourceAPI(with_metaclass(ABCMeta, object)):
    """Class that provide APIs to operate the resource."""

    BACKEND_SERVICE_NAME = None

    def __init__(
        self,
        acs_client: Client,
        header: Optional[Dict[str, str]] = None,
        runtime: Optional[RuntimeOptions] = None,
    ):
        """Initialize a ResourceAPI object.

        Args:
            acs_client (Client): A basic client used to communicate with a specific PAI
                service.
            header (Dict[str, str], optional): Header set in the HTTP request. Defaults
                to None.
            runtime (RuntimeOptions, optional): Options configured for the client
                runtime behavior, such as read_timeout, connection_timeout, etc.
                Defaults to None.
        """
        self.acs_client = acs_client
        self.header = header
        self.runtime = runtime

    def _make_extra_request_options(self):
        """Returns headers and runtime for client."""
        return self.header or dict(), self.runtime or RuntimeOptions()

    def _do_request(self, method_: str, *args, **kwargs):
        headers, runtime = self._make_extra_request_options()
        if "headers" not in kwargs:
            kwargs["headers"] = headers
        if "runtime" not in kwargs:
            kwargs["runtime"] = runtime
        request_method = getattr(self.acs_client, method_)

        return request_method(*args, **kwargs).body

    def get_api_object_by_resource_id(self, resource_id):
        raise NotImplementedError

    def refresh_entity(self, id_, entity):
        """Refresh entity using API object from service."""
        if not isinstance(id_, six.string_types) and not isinstance(
            id_, six.integer_types
        ):
            raise ValueError(
                "Expected integer type or string type for id, but given type %s"
                % type(id_)
            )
        api_obj = self.get_api_object_by_resource_id(resource_id=id_)
        return entity.patch_from_api_object(api_obj)

    @classmethod
    def make_paginated_result(
        cls,
        data: Union[Dict[str, Any], TeaModel],
        item_key=None,
    ) -> "PaginatedResult":
        """Make a paginated result from response.

        Args:
            data: Response data.
            item_key:

        Returns:

        """
        if isinstance(data, TeaModel):
            data = data.to_map()
        total_count = data.pop("TotalCount")

        if item_key:
            items = data[item_key]
        else:
            values = list([val for val in data.values() if isinstance(val, list)])
            if len(values) != 1:
                raise ValueError("Requires item key to make paginated result.")
            items = values[0]
        return PaginatedResult(items=items, total_count=total_count)


class WorkspaceScopedResourceAPI(with_metaclass(ABCMeta, ResourceAPI)):
    """Workspace Scoped Resource API."""

    # A workspace_id placeholder indicate the workspace_id field of
    # the request should not be replaced.
    workspace_id_none_placeholder = "WORKSPACE_ID_NONE_PLACEHOLDER"

    # Default parameter name for request object.
    default_param_name_for_request = "request"

    def __init__(self, workspace_id, acs_client, **kwargs):
        super(WorkspaceScopedResourceAPI, self).__init__(
            acs_client=acs_client, **kwargs
        )
        self.workspace_id = workspace_id

    def _do_request(self, method_, **kwargs):
        request = kwargs.get(self.default_param_name_for_request)

        if not request:
            # Sometimes, request object is not named as "request", we need to find it.
            for param_name, param_value in kwargs.items():
                if isinstance(param_value, TeaModel) and type(
                    param_value
                ).__name__.endswith("Request"):
                    request = param_value
                    break

        # Automatically configure the workspace ID for the request
        if request and hasattr(request, "workspace_id"):
            if request.workspace_id is None:
                request.workspace_id = self.workspace_id
            elif (
                request.workspace_id == self.workspace_id_none_placeholder
                or not request.workspace_id
            ):
                # request.workspace_id is 0 or request.workspace_id is empty string,
                # we do not inject workspace_id of the scope.
                request.workspace_id = None
        return super(WorkspaceScopedResourceAPI, self)._do_request(method_, **kwargs)


class PaginatedResult(object):
    """A class represent response of a pagination call to PAI service."""

    items: List[Union[Dict[str, Any], str]] = None
    total_count: int = None

    def __init__(self, items: List[Union[Dict[str, Any], str]], total_count: int):
        self.items = items
        self.total_count = total_count



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import absolute_import

import glob
import os.path
import pathlib
import tarfile
import tempfile
from typing import Optional, Tuple, Union
from urllib.parse import parse_qs, urlparse

import oss2
from alibabacloud_credentials.client import Client as CredentialClient
from alibabacloud_credentials.models import Config as CredentialConfig
from oss2.credentials import Credentials, CredentialsProvider
from tqdm.autonotebook import tqdm

from .logging import get_logger

logger = get_logger(__name__)


class _ProgressCallbackTqdm(tqdm):
    def __call__(self, consumed_bytes, total_bytes):
        self.update(n=consumed_bytes - self.n)


def _upload_with_progress(
    filename,
    object_key,
    oss_bucket: oss2.Bucket,
):
    local_file_size = os.path.getsize(filename)
    with _ProgressCallbackTqdm(
        total=local_file_size,
        unit="B",
        unit_scale=True,
        desc=f"Uploading file: {filename}",
    ) as pbar:
        oss2.resumable_upload(
            bucket=oss_bucket,
            key=object_key,
            filename=filename,
            progress_callback=pbar,
            num_threads=os.cpu_count() * 2,
        )
        # Mark the progress as completed.
        pbar.update(n=pbar.total - pbar.n)


def _download_with_progress(
    filename,
    object_key,
    oss_bucket: oss2.Bucket,
):
    total = oss_bucket.get_object_meta(object_key).content_length
    with _ProgressCallbackTqdm(
        total=total,
        unit="B",
        unit_scale=True,
        desc=f"Downloading file: {filename}",
    ) as pbar:
        oss2.resumable_download(
            bucket=oss_bucket,
            key=object_key,
            filename=filename,
            progress_callback=pbar,
            num_threads=os.cpu_count(),
        )
        # Mark the progress as completed.
        pbar.update(n=pbar.total - pbar.n)


def is_oss_uri(uri: Union[str, bytes]) -> bool:
    """Determines whether the given uri is an OSS uri.

    Args:
        uri (Union[str, bytes]): A string in OSS URI schema:
            oss://<bucket_name>[.endpoint]/<path/to/file>,


    Returns:
        bool: True if the given uri is an OSS uri, else False.

    """
    return bool(uri and isinstance(uri, (str, bytes)) and str(uri).startswith("oss://"))


class OssUriObj(object):
    """A class that represents an OSS URI and provides some convenient methods."""

    def __init__(self, uri: str):
        """Constructor for class OssUriObj.

        Args:
            uri (str): A string in OSS URI schema: oss://<bucket_name>[.endpoint]/<path/to/file>,
                endpoint in uri is optional.
        """
        if not uri.startswith("oss://"):
            raise ValueError(
                "Invalid OSS URI schema, please provide a string starts with 'oss://'"
            )
        bucket_name, object_key, endpoint, role_arn = self.parse(uri)
        self.bucket_name = bucket_name
        self.object_key = object_key
        self.endpoint = endpoint
        self.role_arn = role_arn

    @classmethod
    def from_bucket_key_endpoint(
        cls, bucket_name: str, object_key: str, endpoint: Optional[str] = None
    ) -> "OssUriObj":
        """Initialize an OSSUri object from bucket_name, object_key and endpoint.

        Args:
            bucket_name (str): The name of the OSS bucket.
            object_key (str): OSS object key/path.
            endpoint (str, optional): Endpoint for the OSS bucket.

        Returns:
            OssUriObj: An OssUriObj instance represents the specified OSS object.

        """
        # OSS object key could not contain leading slashes.
        # Document: https://help.aliyun.com/document_detail/273129.html
        if object_key.startswith("/"):
            logger.warning(
                "OSS object key should not contain leading slashes, the leading"
                " slashes will be removed."
            )
            object_key = object_key.lstrip("/")

        if endpoint:
            if endpoint.startswith("http://"):
                endpoint = endpoint.lstrip("http://")
            elif endpoint.startswith("https://"):
                endpoint = endpoint.lstrip("https://")

            uri = f"oss://{bucket_name}.{endpoint}/{object_key}"
        else:
            uri = f"oss://{bucket_name}/{object_key}"
        return OssUriObj(uri=uri)

    @classmethod
    def parse(cls, oss_uri: str) -> Tuple[str, str, str, str]:
        """Parse OSS uri string and returns a tuple of (bucket_name, object_key,
        endpoint, role_arn).

        Args:
            oss_uri (str): A string in OSS Uri schema: oss://{bucket_name}.{endpoint}/{object_key}.

        Returns:
            Tuple: An tuple of [bucket_name, object_key, endpoint, role_arn].

        """
        parsed_result = urlparse(oss_uri)
        if parsed_result.scheme != "oss":
            raise ValueError(
                "require OSS uri('oss://[bucket_name]/[object_key]') but "
                "given '{}'".format(oss_uri)
            )
        object_key = parsed_result.path
        if object_key.startswith("/"):
            object_key = object_key[1:]

        query = parse_qs(parsed_result.query)
        if "." in parsed_result.hostname:
            bucket_name, endpoint = parsed_result.hostname.split(".", 1)
        else:
            bucket_name = parsed_result.hostname
            # try to get OSS endpoint from url query.
            if "endpoint" in query:
                endpoint = query.get("endpoint")[0]
            elif "host" in query:
                endpoint = query.get("host")[0]
            else:
                endpoint = None
        role_arn = query.get("role_arn")[0] if "role_arn" in query else None

        return bucket_name, object_key, endpoint, role_arn

    def get_uri_with_endpoint(self, endpoint: str = None) -> str:
        """Get an OSS uri string contains endpoint.

        Args:
            endpoint (str): Endpoint of the OSS bucket.

        Returns:
            str: An string in OSS uri schema contains endpoint.

        """
        if not endpoint and not self.endpoint:
            raise ValueError("Unknown endpoint for the OSS bucket.")

        return "oss://{bucket_name}.{endpoint}/{object_key}".format(
            bucket_name=self.bucket_name,
            endpoint=endpoint or self.endpoint,
            object_key=self.object_key,
        )

    def get_dir_uri(self):
        """Returns directory in OSS uri string format of the original object."""
        _, dirname, _ = self.parse_object_key()
        dir_uri = f"oss://{self.bucket_name}{dirname}"
        return dir_uri

    @property
    def uri(self) -> str:
        """Returns OSS uri in string format."""
        return "oss://{bucket_name}/{object_key}".format(
            bucket_name=self.bucket_name,
            object_key=self.object_key,
        )

    def parse_object_key(self) -> Tuple[bool, str, str]:
        """Parse the OSS URI object key, returns a tuple of (is_dir, dir_path, file_name).

        Returns:
            namedtuple: An tuple of is_dir, dir_path, file_name.
        """
        object_key = self.object_key.strip()
        if object_key.endswith("/"):
            is_dir, dir_path, file_name = True, os.path.join("/", object_key), None
        else:
            idx = object_key.rfind("/")
            if idx < 0:
                is_dir, dir_path, file_name = False, "/", object_key
            else:
                is_dir, dir_path, file_name = (
                    False,
                    os.path.join("/", object_key[: idx + 1]),
                    object_key[idx + 1 :],
                )
        return is_dir, dir_path, file_name


def _tar_file(source_file, target=None):
    source_file = (
        source_file if os.path.isabs(source_file) else os.path.abspath(source_file)
    )
    if not os.path.exists(source_file):
        raise ValueError("source file not exists: %s", source_file)
    if os.path.isdir(source_file):
        arcname = ""
    else:
        arcname = os.path.basename(source_file)

    if not target:
        target = tempfile.mktemp()
    with tarfile.open(target, "w:gz") as tar:
        tar.add(name=source_file, arcname=arcname)
    return target


def _get_bucket_and_path(
    bucket: Optional[oss2.Bucket],
    oss_path: Union[str, OssUriObj],
) -> Tuple[oss2.Bucket, str]:
    from pai.session import get_default_session

    sess = get_default_session()
    if isinstance(oss_path, OssUriObj) or is_oss_uri(oss_path):
        # If the parameter oss_path is an OssUriObj object, we need to use the
        # corresponding bucket that OssUriObj instance belongs to for
        # uploading/downloading the data.
        if is_oss_uri(oss_path):
            oss_path = OssUriObj(oss_path)

        if sess.oss_bucket.bucket_name == oss_path.bucket_name:
            bucket = sess.oss_bucket
        else:
            bucket = sess.get_oss_bucket(
                oss_path.bucket_name, endpoint=oss_path.endpoint
            )
        oss_path = oss_path.object_key
    elif not bucket:
        bucket = sess.oss_bucket
    return bucket, oss_path


def upload(
    source_path: str,
    oss_path: Union[str, OssUriObj],
    bucket: Optional[oss2.Bucket] = None,
    is_tar: Optional[bool] = False,
) -> str:
    """Upload local source file/directory to OSS.

    Examples::

        # compress and upload local directory `./src/` to OSS
        >>> upload(source_path="./src/", oss_path="path/to/file",
        ... bucket=session.oss_bucket, is_tar=True)


    Args:
        source_path (str): Source file local path which needs to be uploaded, can be
            a single file or a directory.
        oss_path (Union[str, OssUriObj]): Destination OSS path.
        bucket (oss2.Bucket): OSS bucket used to store the upload data. If it is not
            provided, OSS bucket of the default session will be used.
        is_tar (bool): Whether to compress the file before uploading (default: False).

    Returns:
        str: A string in OSS URI format. If the source_path is directory, return the
            OSS URI representing the directory for uploaded data, else then
            returns the OSS URI points to the uploaded file.
    """

    bucket, oss_path = _get_bucket_and_path(bucket, oss_path)

    source_path_obj = pathlib.Path(source_path)
    if not source_path_obj.exists():
        raise RuntimeError("Source path is not exist: {}".format(source_path))

    if is_tar:
        # compress the local data and upload the compressed source data.
        with tempfile.TemporaryDirectory() as dir_name:
            temp_tar_path = _tar_file(
                source_path, os.path.join(dir_name, "source.tar.gz")
            )
            dest_path = (
                os.path.join(oss_path, os.path.basename(temp_tar_path))
                if oss_path.endswith("/")
                else oss_path
            )
            _upload_with_progress(
                filename=temp_tar_path, object_key=dest_path, oss_bucket=bucket
            )
            return "oss://{}/{}".format(bucket.bucket_name, dest_path)
    elif not source_path_obj.is_dir():
        # if source path is a file, just invoke bucket.put_object.

        # if the oss_path is endswith slash, the file will be uploaded to
        # "{oss_path}{filename}", else the file will be uploaded to "{oss_path}".
        dest_path = (
            os.path.join(oss_path, os.path.basename(source_path))
            if oss_path.endswith("/")
            else oss_path
        )
        _upload_with_progress(
            filename=source_path, object_key=dest_path, oss_bucket=bucket
        )
        return "oss://{}/{}".format(bucket.bucket_name, dest_path)
    else:
        # if the source path is a directory, upload all the file under the directory.
        source_files = glob.glob(
            pathname=str(source_path_obj / "**"),
            recursive=True,
        )
        if not oss_path.endswith("/"):
            oss_path += "/"

        files = [f for f in source_files if not os.path.isdir(f)]
        for file_path in files:
            file_path_obj = pathlib.Path(file_path)
            file_relative_path = file_path_obj.relative_to(source_path_obj).as_posix()
            object_key = oss_path + file_relative_path
            _upload_with_progress(
                filename=file_path, object_key=object_key, oss_bucket=bucket
            )
        return "oss://{}/{}".format(bucket.bucket_name, oss_path)


def download(
    oss_path: Union[str, OssUriObj],
    local_path: str,
    bucket: Optional[oss2.Bucket] = None,
    un_tar=False,
):
    """Download OSS objects to local path.

    Args:
        oss_path (str): Source OSS path, could be a single OSS object or a OSS
            directory.
        local_path (str): Local path used to store the data from OSS.
        bucket (oss2.Bucket, optional): OSS bucket used to store the upload data. If it
            is not provided, OSS bucket of the default session will be used.
        un_tar (bool, optional): Whether to decompress the downloaded data. It is only
            work for `oss_path` point to a single file that has a suffix "tar.gz".

    Returns:
        str: A local file path for the downloaded data.

    """

    bucket, oss_path = _get_bucket_and_path(bucket, oss_path)

    if not bucket.object_exists(oss_path) or oss_path.endswith("/"):
        # The `oss_path` represents a "directory" in the OSS bucket, download the
        # objects which object key is prefixed with `oss_path`.
        # Note: `un_tar` is not work while `oss_path` is a directory.

        oss_path += "/" if not oss_path.endswith("/") else ""
        iterator = oss2.ObjectIteratorV2(
            bucket=bucket,
            prefix=oss_path,
        )
        keys = [obj.key for obj in iterator if not obj.key.endswith("/")]
        for key in tqdm(keys, desc=f"Downloading: {oss_path}"):
            rel_path = os.path.relpath(key, oss_path)
            dest = os.path.join(local_path, rel_path)
            os.makedirs(os.path.dirname(dest), exist_ok=True)
            _download_with_progress(
                dest,
                object_key=key,
                oss_bucket=bucket,
            )
        return local_path
    else:
        # The `oss_path` represents a single file in OSS bucket.
        if oss_path.endswith(".tar.gz") and un_tar:
            # currently, only tar.gz format is supported for un_tar after downloading.
            with tempfile.TemporaryDirectory() as temp_dir:
                target_path = os.path.join(temp_dir, os.path.basename(oss_path))
                _download_with_progress(
                    target_path,
                    object_key=oss_path,
                    oss_bucket=bucket,
                )
                with tarfile.open(name=target_path, mode="r") as t:
                    t.extractall(path=local_path)

            return local_path
        else:
            os.makedirs(local_path, exist_ok=True)
            dest = os.path.join(local_path, os.path.basename(oss_path))
            _download_with_progress(
                dest,
                object_key=oss_path,
                oss_bucket=bucket,
            )

            return dest


class CredentialProviderWrapper(CredentialsProvider):
    """A wrapper class for the credential provider of OSS."""

    def __init__(self, config: Union[CredentialConfig] = None):
        self.client = CredentialClient(config)

    def get_credentials(self) -> Credentials:
        return Credentials(
            access_key_id=self.client.get_access_key_id(),
            access_key_secret=self.client.get_access_key_secret(),
            security_token=self.client.get_security_token(),
        )



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import urllib.request
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.error import HTTPError

import backoff
import numpy
import numpy as np
import six
from eas_prediction import pytorch_predict_pb2 as pt_pb
from eas_prediction import tf_request_pb2 as tf_pb

from .common.logging import get_logger
from .session import Session, get_default_session

logger = get_logger(__name__)


def _is_pil_image(data) -> bool:
    try:
        from PIL import Image

        return isinstance(data, Image.Image)
    except ImportError:
        return False


def _is_numpy_ndarray(data) -> bool:
    try:
        import numpy

        return isinstance(data, numpy.ndarray)
    except ImportError:
        return False


def _is_pandas_dataframe(data) -> bool:
    try:
        import pandas

        return isinstance(data, pandas.DataFrame)
    except ImportError:
        return False


class TensorFlowIOSpec(object):
    def __init__(self, name: str, shape: Tuple, data_type: tf_pb.ArrayDataType):
        """A class represents TensorFlow inputs/outputs spec.

        Args:
            name (str): Name for the spec.
            shape (Tuple): The shape of the input/output value.
            data_type (tf_pb.ArrayDataType): Data type of the value.
        """
        self.name = name
        self.shape = shape
        self.data_type = data_type


class SerializerBase(ABC):
    """Abstract class for creating a Serializer class for predictor."""

    @abstractmethod
    def serialize(self, data) -> bytes:
        """Serialize the input data to bytes for transmitting."""

    @abstractmethod
    def deserialize(self, data: bytes):
        """Deserialize the data from raw bytes to Python object ."""

    def inspect_from_service(
        self, service_name: str, *, session: Optional[Session] = None
    ):
        """Inspect the online prediction service to complete the serializer instance
        initialization.

        The implementation of the `inspect_from_service` method is optional. You only
        need to implement it if your serializer requires additional information from
        service metadata or if it needs to send a request to the service in order to
        be initialized.

        """


class BytesSerializer(SerializerBase):
    """A Serializer object that serialize input data into bytes format and deserialize
    bytes formatted data into python object."""

    def serialize(self, data) -> bytes:
        if isinstance(data, (dict, list, tuple)):
            return json.dumps(data).encode()
        elif isinstance(data, str):
            return data.encode()
        elif isinstance(data, bytes):
            return data
        else:
            return str(data).encode()

    def deserialize(self, data: bytes):
        return data


class JsonSerializer(SerializerBase):
    """A Serializer object that serialize input data into JSON format and deserialize
    JSON formatted data into python object."""

    def serialize(self, data) -> bytes:
        if isinstance(data, six.string_types):
            return data

        if _is_pandas_dataframe(data):
            data = data.to_numpy().tolist()
        elif _is_numpy_ndarray(data):
            data = data.tolist()
        return json.dumps(data).encode()

    def deserialize(self, data):
        return json.loads(data)


class TensorFlowSerializer(SerializerBase):
    """A Serializer class that responsible for transforming input/output data for
    TensorFlow processor service."""

    NUMPY_DATA_TYPE_MAPPING = {
        "DT_FLOAT": np.float32,
        "DT_DOUBLE": np.float64,
        "DT_INT8": np.int8,
        "DT_INT16": np.int16,
        "DT_INT32": np.int32,
        "DT_INT64": np.int64,
        "DT_UINT8": np.uint8,
        "DT_UINT16": np.uint16,
        "DT_BOOL": np.bool_,
        "DT_STRING": np.str_,
    }

    def __init__(
        self,
    ):
        """TensorflowSerializer initializer."""

        self._input_specs = []
        self._output_filter = []
        self._signature_name = None
        super(TensorFlowSerializer, self).__init__()

    def inspect_from_service(
        self, service_name: str, *, session: Optional[Session] = None
    ):
        """Inspect the service to complete serializer instance initialization.

        Args:
            service_name (str): Name of the online prediction service.
            session (:class:`pai.session.Session`):  A PAI session instance used for
                communicating with PAI services.

        """
        session = session or get_default_session()
        sig_def = self.inspect_model_signature_def(service_name, session=session)
        self._init_from_signature_def(sig_def)

    @classmethod
    def inspect_model_signature_def(
        cls, service_name: str, *, session: Session = None
    ) -> Dict[str, Any]:
        """Inspect the TensorFlow serving model signature by sending a request to
        the service.

        TensorFlow processor creates a prediction service and exposes an HTTP API for
        model signature definition.

        Example API returns::

            {
                "signature_name": "serving_default",
                "inputs": [
                    {
                        "name": "flatten_input",
                        "shape": [
                            -1,
                            28,
                            28
                        ],
                        "type": "DT_FLOAT"
                    }
                ],
                "outputs": [
                    {
                        "name": "dense_1",
                        "shape": [
                            -1,
                            10
                        ],
                        "type": "DT_FLOAT"
                    }
                ]
            }

        Returns:
            A dictionary that represents the model signature definition.

        """
        from pai.predictor import ServiceStatus

        session = session or get_default_session()

        service_api_object = session.service_api.get(service_name)
        if service_api_object["Status"] != ServiceStatus.Running:
            raise RuntimeError(
                f"Service is not ready, cannot send request to the service to inspect "
                f"model signature definition: "
                f"name={service_api_object['ServiceName']} "
                f"status={service_api_object['Status']} "
                f"reason={service_api_object['Reason']} "
                f"message={service_api_object['Message']}."
            )

        @backoff.on_exception(
            backoff.expo,
            exception=HTTPError,
            max_tries=3,
            max_time=10,
        )
        def _send_request():
            request = urllib.request.Request(
                url=service_api_object["InternetEndpoint"],
                headers={
                    "Authorization": service_api_object["AccessToken"],
                },
            )
            resp = urllib.request.urlopen(request)
            return resp

        resp = _send_request()
        signature_def = json.load(resp)
        return signature_def

    def serialize(self, data: Union[Dict[str, Any], tf_pb.PredictRequest]) -> bytes:

        if isinstance(data, tf_pb.PredictRequest):
            return data.SerializeToString()

        request = tf_pb.PredictRequest()
        if self._output_filter:
            for output_name in self._output_filter:
                request.output_filter.append(output_name)

        if not isinstance(data, dict):
            if not self._input_specs or len(self._input_specs) > 1:
                raise ValueError(
                    "TensorFlowSerializer accepts a dictionary as input data, "
                    "with each input value having a name."
                )
            else:
                # TensorFlow Processor expects key-value pairs for input data. However,
                # if the input data is not a dictionary and the deployed model accepts
                # exactly one input (by model signature), the key will be inferred
                # from model signature and the current input data will be taken as
                # value.
                value = numpy.asarray(data)
                input_spec = self._input_specs[0]
                if (
                    input_spec.shape
                    and len([dim for dim in input_spec.shape if dim == -1]) == 1
                ):
                    value = value.reshape(input_spec.shape)
                data_type = (
                    input_spec.data_type
                    if input_spec and input_spec.data_type is not None
                    else self._np_dtype_to_tf_dtype(value.dtype.type)
                )
                self._put_value(
                    request=request,
                    name=input_spec.name,
                    data_type=data_type,
                    shape=value.shape,
                    data=np.ravel(value).tolist(),
                )
        else:
            input_specs_dict = (
                {input_spec.name: input_spec for input_spec in self._input_specs}
                if self._input_specs
                else {}
            )
            for name, value in data.items():
                input_spec = input_specs_dict.get(name)
                if not isinstance(value, np.ndarray):
                    value = np.asarray(value)
                data_type = (
                    input_spec.data_type
                    if input_spec and input_spec.data_type is not None
                    else self._np_dtype_to_tf_dtype(value.dtype.type)
                )

                if (
                    input_spec
                    and input_spec.shape
                    and len([dim for dim in input_spec.shape if dim == -1]) == 1
                ):
                    value = value.reshape(input_spec.shape)
                self._put_value(
                    request=request,
                    name=input_spec.name,
                    data_type=data_type,
                    shape=value.shape,
                    data=np.ravel(value).tolist(),
                )

        return request.SerializeToString()

    def _init_from_signature_def(self, signature_def):
        """Build TensorFlowSerializer from signature def.

        Args:
            signature_def: Signature def returns from PAI-EAS tensorflow processor.

        Returns:
            TensorFlowSerializer:
        """
        inputs = signature_def["inputs"]
        signature_def_key = signature_def["signature_name"]
        input_specs = []
        output_specs = []
        for input_def in inputs:
            data_type = tf_pb.ArrayDataType.Value(input_def["type"])
            input_spec = TensorFlowIOSpec(
                name=input_def["name"],
                data_type=data_type,
                # use batch_size=1
                shape=input_def["shape"][1:],
            )
            input_specs.append(input_spec)

        for output_def in signature_def["outputs"]:
            data_type = tf_pb.ArrayDataType.Value(output_def["type"])
            output_spec = TensorFlowIOSpec(
                name=output_def["name"],
                data_type=data_type,
                shape=output_def["shape"],
            )
            output_specs.append(output_spec)

        if not self._signature_name:
            self._signature_name = signature_def_key

        if not self._input_specs:
            self._input_specs = input_specs
        if not self._output_filter:
            self._output_filter = [spec.name for spec in output_specs]

    def deserialize(self, data: bytes):
        response = tf_pb.PredictResponse()
        response.ParseFromString(data)
        output_names = response.outputs.keys()
        results = {}
        for name in output_names:
            results[name] = self._get_value(
                response=response,
                name=name,
            )
        return results

    def _np_dtype_to_tf_dtype(self, np_dtype):
        rev_map = {value: key for key, value in self.NUMPY_DATA_TYPE_MAPPING.items()}
        if np_dtype not in rev_map:
            raise ValueError(
                f"Numpy dtype {np_dtype} is not supported in TensorFlowSerializer."
            )

        return tf_pb.ArrayDataType.Value(rev_map[np_dtype])

    def _tf_dtype_to_np_dtype(self, data_type):
        data_type_name = tf_pb.ArrayDataType.Name(data_type)
        if data_type_name not in self.NUMPY_DATA_TYPE_MAPPING:
            raise ValueError(
                f"Data type {data_type_name} is not supported in TensorFlowSerializer."
            )
        return self.NUMPY_DATA_TYPE_MAPPING.get(data_type_name)

    def _put_value(
        self, request: tf_pb.PredictRequest, name: str, data_type, shape, data
    ):
        request.inputs[name].dtype = data_type
        request.inputs[name].array_shape.dim.extend(shape)

        integer_types = [
            tf_pb.DT_INT8,
            tf_pb.DT_INT16,
            tf_pb.DT_INT32,
            tf_pb.DT_UINT8,
            tf_pb.DT_UINT16,
            tf_pb.DT_QINT8,
            tf_pb.DT_QINT16,
            tf_pb.DT_QINT32,
            tf_pb.DT_QUINT8,
            tf_pb.DT_QUINT16,
        ]
        if data_type == tf_pb.DT_FLOAT:
            request.inputs[name].float_val.extend(data)
        elif data_type == tf_pb.DT_DOUBLE:
            request.inputs[name].double_val.extend(data)
        elif data_type in integer_types:
            request.inputs[name].int_val.extend(data)
        elif data_type == tf_pb.DT_INT64:
            request.inputs[name].int64_val.extend(data)
        elif data_type == tf_pb.DT_BOOL:
            request.inputs[name].bool_val.extend(data)
        elif data_type == tf_pb.DT_STRING:
            request.inputs[name].string_val.extend(data)
        else:
            raise ValueError(
                f"Not supported input data type for TensorFlow PredictRequest: {data_type}"
            )

    def _get_value(self, response: tf_pb.PredictResponse, name):
        output = response.outputs[name]

        if (
            name not in response.outputs
            or tf_pb.DT_INVALID == response.outputs[name].dtype
        ):
            return
        np_dtype = self._tf_dtype_to_np_dtype(response.outputs[name].dtype)
        shape = list(output.array_shape.dim)

        if output.dtype == tf_pb.DT_FLOAT:
            return np.asarray(output.float_val, np_dtype).reshape(shape)
        elif output.dtype in (tf_pb.DT_INT8, tf_pb.DT_INT16, tf_pb.DT_INT32):
            return np.asarray(output.int_val, np_dtype).reshape(shape)
        elif output.dtype == tf_pb.DT_INT64:
            return np.asarray(output.int64_val, np_dtype).reshape(shape)
        elif output.dtype == tf_pb.DT_DOUBLE:
            return np.asarray(output.double_val, np_dtype).reshape(shape)
        elif output.dtype == tf_pb.DT_STRING:
            return np.asarray(output.string_val, np_dtype).reshape(shape)
        elif output.dtype == tf_pb.DT_BOOL:
            return np.asarray(output.bool_val, np_dtype).reshape(shape)
        else:
            raise ValueError(f"Not support data_type: {output.dtype}")


class PyTorchSerializer(SerializerBase):
    """A serializer responsible for transforming input/output data for PyTorch
    processor service.

    """

    NUMPY_DATA_TYPE_MAPPING = {
        "DT_FLOAT": np.float32,
        "DT_DOUBLE": np.float64,
        "DT_INT8": np.int8,
        "DT_INT16": np.int16,
        "DT_INT32": np.int32,
        "DT_INT64": np.int64,
        "DT_UINT8": np.uint8,
        "DT_UINT16": np.uint16,
        "DT_BOOL": np.bool_,
        "DT_STRING": np.str_,
    }

    def __init__(
        self,
    ):
        self._output_filter = []

    def _np_dtype_to_torch_dtype(self, np_dtype):
        """Get PredictRequest data_type from dtype of input np.ndarray."""
        rev_map = {value: key for key, value in self.NUMPY_DATA_TYPE_MAPPING.items()}
        if np_dtype not in rev_map:
            raise ValueError(
                f"Numpy dtype {np_dtype} is not supported in PyTorchSerializer."
            )
        return pt_pb.ArrayDataType.Value(rev_map[np_dtype])

    def _torch_dtype_to_numpy_dtype(self, data_type):
        data_type_name = pt_pb.ArrayDataType.Name(data_type)
        if data_type_name not in self.NUMPY_DATA_TYPE_MAPPING:
            raise ValueError(
                f"Data type {data_type_name} is not supported in PyTorchSerializer."
            )
        return self.NUMPY_DATA_TYPE_MAPPING.get(data_type_name)

    def serialize(self, data: Union[np.ndarray, List, Tuple]) -> bytes:
        request = pt_pb.PredictRequest()
        if _is_pil_image(data):
            data = np.asarray(data)
        elif isinstance(data, (bytes, str)):
            data = np.asarray(data)

        if isinstance(data, np.ndarray):
            # if input data type is np.ndarray, we assume there is only one input data
            # for the prediction request.
            self._put_value(
                request,
                index=0,
                shape=data.shape,
                data_type=self._np_dtype_to_torch_dtype(data.dtype.type),
                data=np.ravel(data).tolist(),
            )
        elif isinstance(data, (List, Tuple)):
            # if input data type is List or Tuple, we assume there is multi input data.
            # for the prediction request.
            for idx, item in enumerate(data):
                if not isinstance(item, np.ndarray):
                    item = np.asarray(item)
                if not item:
                    continue
                self._put_value(
                    request,
                    index=0,
                    shape=item.shape,
                    data_type=self._np_dtype_to_torch_dtype(item.dtype.type),
                    data=np.ravel(item).tolist(),
                )
        else:
            raise ValueError(
                "PyTorchSerializer accept List, Tuple as input request data."
            )
        return request.SerializeToString()

    def deserialize(self, data: bytes):
        resp = pt_pb.PredictResponse()
        resp.ParseFromString(data)
        if len(resp.outputs) > 1:
            results = []
            for idx in range(resp.outputs):
                results.append(self._get_value(resp, idx))
            return results
        elif len(resp.outputs) == 1:
            return self._get_value(resp, index=0)

    def _put_value(
        self, request: pt_pb.PredictRequest, index: int, shape, data_type, data
    ):
        while len(request.inputs) < index + 1:
            request.inputs.add()
        request.inputs[index].dtype = data_type
        request.inputs[index].array_shape.dim.extend(shape)
        if data_type == pt_pb.DT_FLOAT:
            request.inputs[index].float_val.extend(data)
        elif data_type == pt_pb.DT_DOUBLE:
            request.inputs[index].double_val.extend(data)
        elif data_type in (
            pt_pb.DT_INT8,
            pt_pb.DT_INT16,
            pt_pb.DT_INT32,
            pt_pb.DT_UINT8,
        ):
            request.inputs[index].int_val.extend(data)
        elif data_type == pt_pb.DT_INT64:
            request.inputs[index].int64_val.extend(data)
        else:
            raise ValueError(f"Not supported PyTorch request data type: {data_type}")

    def _get_value(self, response: pt_pb.PredictResponse, index: int):
        output = response.outputs[index]
        if output.dtype == pt_pb.DT_INVALID:
            return

        np_dtype = self._torch_dtype_to_numpy_dtype(output.dtype)
        shape = list(output.array_shape.dim)
        if output.dtype == pt_pb.DT_FLOAT:
            return np.asarray(output.float_val, np_dtype).reshape(shape)
        elif output.dtype in (
            pt_pb.DT_INT8,
            pt_pb.DT_INT16,
            pt_pb.DT_INT32,
            pt_pb.DT_UINT8,
        ):
            return np.asarray(output.int_val, np_dtype).reshape(shape)
        elif output.dtype == pt_pb.DT_INT64:
            return np.asarray(output.int64_val, np_dtype).reshape(shape)
        elif output.dtype == pt_pb.DT_DOUBLE:
            return np.asarray(output.double_val, np_dtype).reshape(shape)
        else:
            raise ValueError(
                f"Not supported PyTorch response data type: {output.dtype}"
            )



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..common.yaml_utils import dump as yaml_dump
from ..libs.alibabacloud_paiflow20210202.models import (
    CreateRunRequest,
    GetNodeRequest,
    GetNodeResponseBody,
    GetRunRequest,
    GetRunResponseBody,
    ListNodeLogsRequest,
    ListNodeLogsResponseBody,
    ListNodeOutputsRequest,
    ListNodeOutputsResponseBody,
    ListRunsRequest,
    ListRunsResponseBody,
    UpdateRunRequest,
)
from .base import ServiceName, WorkspaceScopedResourceAPI


class PipelineRunAPI(WorkspaceScopedResourceAPI):

    BACKEND_SERVICE_NAME = ServiceName.PAIFLOW

    _list_method = "list_runs_with_options"
    _get_method = "get_run_with_options"
    _create_method = "create_run_with_options"
    _start_method = "start_run_with_options"
    _terminate_method = "terminate_run_with_options"
    _update_method = "update_run_with_options"

    _get_node_method = "get_node_with_options"
    _list_node_logs_method = "list_node_logs_with_options"
    _list_node_outputs_method = "list_node_outputs_with_options"

    def list(
        self,
        name=None,
        run_id=None,
        pipeline_id=None,
        status=None,
        sort_by=None,
        order=None,
        page_number=None,
        page_size=None,
        experiment_id=None,
        source=None,
        **kwargs,
    ):
        request = ListRunsRequest(
            page_number=page_number,
            page_size=page_size,
            experiment_id=experiment_id,
            name=name,
            pipeline_id=pipeline_id,
            run_id=run_id,
            sort_by=sort_by,
            order=order,
            source=source,
            status=status,
            **kwargs,
        )
        resp: ListRunsResponseBody = self._do_request(
            method_=self._list_method, request=request
        )
        return self.make_paginated_result(resp.to_map())

    def get(self, run_id):
        request = GetRunRequest()
        resp: GetRunResponseBody = self._do_request(
            method_=self._get_method, run_id=run_id, request=request
        )
        return resp.to_map()

    def create(
        self,
        name,
        pipeline_id=None,
        manifest=None,
        arguments=None,
        env=None,
        no_confirm_required=False,
        source="SDK",
    ):
        run_args = {"arguments": arguments, "env": env}

        if not pipeline_id and not manifest:
            raise ValueError(
                "Create pipeline run require either pipeline_id or manifest."
            )
        if pipeline_id and manifest:
            raise ValueError(
                "Both pipeline_id and manifest are provide, create_run need only one."
            )
        if not name:
            raise ValueError("Pipeline run instance need a name.")

        run_args = yaml_dump(run_args)
        if manifest and isinstance(manifest, dict):
            manifest = yaml_dump(manifest)

        request = CreateRunRequest(
            pipeline_id=pipeline_id,
            name=name,
            pipeline_manifest=manifest,
            arguments=run_args,
            no_confirm_required=no_confirm_required,
            source=source,
        )
        resp = self._do_request(self._create_method, request=request)
        return resp.run_id

    def start(self, run_id):
        self._do_request(self._start_method, run_id=run_id)

    def terminate_run(self, run_id):
        self._do_request(self._terminate_method, run_id=run_id)

    def update(self, run_id, name):
        request = UpdateRunRequest(name=name)
        self._do_request(self._update_method, run_id=run_id, request=request)

    def get_node(self, run_id, node_id, depth=2):
        request = GetNodeRequest(depth=depth)
        resp: GetNodeResponseBody = self._do_request(
            method_=self._get_node_method,
            run_id=run_id,
            node_id=node_id,
            request=request,
        )
        return resp.to_map()

    def list_node_logs(
        self,
        run_id,
        node_id,
        from_time=None,
        to_time=None,
        keyword=None,
        reverse=False,
        page_offset=0,
        page_size=100,
    ):
        request = ListNodeLogsRequest(
            offset=page_offset,
            page_size=page_size,
            from_time_in_seconds=from_time,
            to_time_in_seconds=to_time,
            keyword=keyword,
            reverse=reverse,
        )
        resp: ListNodeLogsResponseBody = self._do_request(
            self._list_node_logs_method,
            run_id=run_id,
            node_id=node_id,
            request=request,
        )
        return self.make_paginated_result(resp.to_map())

    def list_node_outputs(
        self,
        run_id,
        node_id,
        depth=2,
        name=None,
        sort_by=None,
        order=None,
        type=None,
        page_number=1,
        page_size=50,
    ):

        request = ListNodeOutputsRequest(
            name=name,
            depth=depth,
            page_number=page_number,
            page_size=page_size,
            sort_by=sort_by,
            order=order,
            type=type,
        )

        resp: ListNodeOutputsResponseBody = self._do_request(
            self._list_node_outputs_method,
            run_id=run_id,
            node_id=node_id,
            request=request,
        )
        return self.make_paginated_result(resp.to_map())



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import absolute_import

from typing import Optional

from alibabacloud_credentials.client import Client as CredentialClient
from alibabacloud_sts20150401.client import Client as StsClient
from alibabacloud_tea_openapi.models import Config

from ..common.consts import Network
from ..common.logging import get_logger
from ..common.utils import http_user_agent
from ..libs.alibabacloud_aiworkspace20210204.client import Client as WorkspaceClient
from ..libs.alibabacloud_eas20210701.client import Client as EasClient
from ..libs.alibabacloud_pai_dlc20201203.client import Client as DlcClient
from ..libs.alibabacloud_pai_dsw20220101.client import Client as DswClient
from ..libs.alibabacloud_paiflow20210202.client import Client as FlowClient
from ..libs.alibabacloud_paistudio20220112.client import Client as PaiClient
from .base import ServiceName

_logger = get_logger(__name__)

DEFAULT_SERVICE_ENDPOINT_PATTERN = "{}.{}.aliyuncs.com"


class ClientFactory(object):
    ClientByServiceName = {
        ServiceName.PAI_DLC: DlcClient,
        ServiceName.PAI_EAS: EasClient,
        ServiceName.PAI_WORKSPACE: WorkspaceClient,
        ServiceName.PAIFLOW: FlowClient,
        ServiceName.PAI_STUDIO: PaiClient,
        ServiceName.STS: StsClient,
        ServiceName.PAI_DSW: DswClient,
    }

    @staticmethod
    def _is_inner_client(acs_client):
        return acs_client.get_region_id() == "center"

    @classmethod
    def create_client(
        cls,
        service_name,
        region_id: str,
        credential_client: CredentialClient,
        network: Optional[Network] = None,
        **kwargs,
    ):
        """Create an API client which is responsible to interacted with the Alibaba
        Cloud service."""

        config = Config(
            region_id=region_id,
            credential=credential_client,
            endpoint=cls.get_endpoint(
                service_name=service_name,
                region_id=region_id,
                network=network,
            ),
            signature_algorithm="v2",
            user_agent=http_user_agent(),
            **kwargs,
        )
        client = cls.ClientByServiceName.get(service_name)(config)
        return client

    @classmethod
    def get_endpoint(
        cls, service_name: str, region_id: str, network: Optional[Network] = None
    ) -> str:
        """Get the endpoint for the service client."""
        if not region_id:
            raise ValueError("Please provide region_id to get the endpoint.")

        if network and network != Network.PUBLIC:
            if service_name == "pai-eas":
                # see endpoint list provided by PAI-EAS
                # https://next.api.aliyun.com/product/eas
                subdomain = f"pai-eas-manage-{network.value.lower()}"
            else:
                subdomain = f"{service_name}-{network.value.lower()}"
        else:
            subdomain = service_name
        return DEFAULT_SERVICE_ENDPOINT_PATTERN.format(subdomain, region_id)



from __future__ import absolute_import

import os

from setuptools import find_packages, setup

pkg_root = os.path.dirname(os.path.abspath(__file__))

REQUIREMENTS_FILE = "requirements/requirements.txt"
PACKAGE_NAME = os.getenv("PACKAGE_NAME", "pai")

version_data = {}
with open(os.path.join(pkg_root, "pai/version.py")) as fp:
    exec(fp.read(), version_data)
version = version_data["VERSION"]


def read_requirements():
    with open(os.path.join(pkg_root, REQUIREMENTS_FILE), "r") as f:
        return [line.strip() for line in f.readlines() if line.strip()]


long_description = None
if os.path.exists("README.md"):
    with open("README.md") as f:
        long_description = f.read()

setup(
    name=PACKAGE_NAME,
    python_requires=">=3.8",
    version=version,
    description="Alibaba Cloud PAI Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    url="https://www.aliyun.com/product/bigdata/product/learn",
    packages=find_packages(include=["pai", "pai.*"]),
    install_requires=read_requirements(),
    author="Alibaba PAI team",
    keywords="ML Alibaba Cloud PAI Training Inference Pipeline",
    license="Apache License 2.0",
    classifier=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
    ],
)
